# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, 2020, Knut-Frode Dagestad, MET Norway

"""
This file is a template for an ocean trajectory model
to be implemented within the OpenDrift framework.
To create a new model, make a copy (renamed) of this file, and
modify the class names and details below, according to the comments.
Alternatively, another existing class can be used as a starting point.

For more help, see https://opendrift.github.io/writing_a_new_model.html
For usage of the new model, see example scripts.

First we define an element type to be used by the trajectory model.
An element type (class) may also be imported/reused from
another module - see examples in the folder "elements"

"""

import numpy as np
from opendrift.models.oceandrift import OceanDrift
from opendrift.models.oceandrift import Lagrangian3DArray

class TemplateElementType(Lagrangian3DArray):
    """Extending LagrangianArray with relevant properties."""

    # Define and name the properties which the elements shall have.
    # These are added to the inherited properties (e.g. from OceanDrift)
    # Property names may be freely chosen, but the "update" function (below)
    # will refer to the names in the specifications of the processes.
    #
    # The "default" value is used for alle elements, unless a custom
    # value is provided at time of seeding:
    # o.seed_elements(......., <property>=<value>)
    #
    # The default value can also be updated by the user
    # before seeding with the config mechanism:
    # o.set_config('seed:<property>', <new default value>)
    #
    # The "units" attribute is used for labeling.

    variables = Lagrangian3DArray.add_variables([
        ('new_property_1', {'dtype': np.float32,
                            'units': '1',
                            'default': 0}),
        ('new_property_2', {'dtype': np.float32,
                            'units': '1',
                            'default': 0})])


class ModelTemplate(OceanDrift):
    """Open source trajectory model based on the OpenDrift framework.

    This is a template model, to be copied and modified according to need.
    This class is a subclass of "OceanDrift", and thus inherits all its
    functionality. More specialised models may also be subclassed.
    """

    # Specify which element/particle type (class) to be used by this model.
    # A model can have/use only one element type
    ElementType = TemplateElementType

    # Specify which environment variables (e.g. wind, waves, currents...)
    # are needed/required by the present model, to be used for updating
    # the element properties (including propagation).
    # The convention of using CF standard_name (https://cfconventions.org)
    # allows reusing readers for many/all models.
    # Vector quantities should be referred to by the x-y components
    # (e.g. x_wind) and not the lon-lat componenets (e.g. east_wind).
    required_variables = {
        'x_sea_water_velocity': {'fallback': 0},
        'y_sea_water_velocity': {'fallback': 0},
        'x_wind': {'fallback': 0, 'important': False},
        'y_wind': {'fallback': 0, 'important': False},
        'ocean_vertical_diffusivity': {'fallback': 0.02, 'important': False, 'profiles': True},
        'land_binary_mask': {'fallback': None},
        }
    # Default (fallback) numerical values will be used for particles 
    # outside of the spatial or tempoaral domain of all the available readers,
    # or if no readers are given for the specific variables.
    # If the fallback value is None, this means that simulation will
    # not run unless a reader has been added to provide this variable.
    # The attribute *important' relates to the concept of Lazy readers:
    # https://opendrift.github.io/tutorial.html#lazy-readers
    # Lazy readers will be initialized successively until there are active
    # readers for all variables, except for variables which are
    # marked as not important, as for wind in the example above.
    # If the attribute 'profiles' is True for a variable (as for
    # ocean_vertical_diffusivity in the example above), a vertical profile
    # of this variable is obtained over the vertical depth interval as 
    # defined below. These profiles are used for the vertical turbulence scheme:
    # https://opendrift.github.io/_modules/opendrift/models/oceandrift.html#OceanDrift.vertical_mixing
    # Profiles are typically required for ocean_vertical_diffusivity, and sometimes for 
    # salinity and temperature in order to calculate the stratification.
    required_profiles_z_range = [-50, 0]
    
    def __init__(self, *args, **kwargs):

        # Here any preprocessing and initialisation of a simulation
        # may be performed, e.g. importing external data from file.

        # The constructor of parent class must always be called
        # to perform some necessary common initialisation tasks:
        super(ModelTemplate, self).__init__(*args, **kwargs)
            
        # The above initialisation includes adding common config options.
        # Further config options specific to this subclass can be
        # added with the following method (example options below from OpenOil):
        self._add_config({
            'seed:m3_per_hour': {'type': 'float', 'default': 1,
                'min': 0, 'max': 1e10, 'units': 'm3 per hour',
                'description': 'The amount (volume) of oil released per hour (or total amount if release is instantaneous)',
                'level': self.CONFIG_LEVEL_ESSENTIAL},
            'seed:droplet_diameter_min_subsea': {'type': 'float', 'default': 0.0005,
                'min': 1e-8, 'max': 1, 'units': 'meters',
                'description': 'The minimum dimaeter of oil droplet for a subsea release.',
                'level': self.CONFIG_LEVEL_BASIC},
            'processes:dispersion': {'type': 'bool', 'default': True,
                'description': 'Oil is removed from simulation (dispersed), if entrained as very small droplets.',
                'level': self.CONFIG_LEVEL_BASIC},
            'wave_entrainment:droplet_size_distribution': {'type': 'enum', 'enum': ['Johansen et al. (2015)', 'Li et al. (2017)'], 'default': 'Johansen et al. (2015)',
                'level': self.CONFIG_LEVEL_ADVANCED, 'description':
                'Algorithm to be used for calculating oil droplet size spectrum after entrainment by breaking waves.'},
           })
        
        # The _add_config method takes a dictionary where keys are config settings,
        # which can be adjusted by the user by:
        # >>> o.set_config(key, value)
        # and whose actual value is obtained during the simulation (*update* method below) by
        # value = o.get_config(key)
        # The config keys have a colon (:) separator for logical grouping into categories.
        # The config specification above has the following items:
        # - type: one of ['float', 'int', 'bool', 'enum']
        # - default: the default value
        # - units: the unit, used for labeling (potentially later for automatic scaling)
        # - description: a human readable help text, e.g. to be displayed in the GUI
        # - level:
        #   - CONFIG_LEVEL_ESSENTIAL (1): for config settings which the user must consider.
        #        These config items are e.g. shown on the front/seeding page of a GUI
        #   - CONFIG_LEVEL_BASIC (2): for config settings which most users may consider to change
        #   - CONFIG_LEVEL_ADVANCED (3): for config setting which only advanced users might change
        #
        # - If 'type' is 'float' or 'int' there are two additional items:
        #    - 'min' and 'max' which are the minimum and maximum allowed values of this setting.
        #       The set_config() method will raise an error of provided values are outside of this range.
        # - If 'type' is 'enum' there is an additional item 'enum' which is a list
        #     of the allowed values. These may be strings or numbers.
        #
        # Dafault values of config-settings which are inherited from parent models
        # may be overriden for the current module by:
        self._set_config_default('drift:vertical_mixing', True)

    def update(self):
        """Update positions and properties of particles."""

        # This function (``update``, may not be renamed) defines
        # the actions (processes) to be performed at each time step.
        # The default timestep is one hour, but this may be overriden
        # by the user.
        #
        # The elements and corresponding environment information
        # (wind, waves, current...) is available to this function as
        # the recarrays ``elements`` and ``environment``, respectively.
        # Element properties and environvent variables are named arrays
        # of same length (number of active particles), or scalars.
        #
        # E.g.:
        #    self.elements.z
        #    self.elements.mass
        #    self.environment.x_wind
        #    self.environment.y_wind
        #
        # E.g. to specify evaporation (mass reduction) proportional to wind speed:
        #
        #    self.elements.mass = self.elements.mass - \
        #        np.sqrt(self.environment.x_wind**2 + \
        #                self.environment.y_wind**2)*.01
        #
        # If particles need to be deactived (e.g. if
        # self.environemtn.someproperty is 0), the special function
        # `self.deactivate_elements()` should be used:
        # `self.deactivate_elements(self.environment.someproperty==0)`
        # 
        # Interaction with coast is handled by ``basemodel`` and is
        # chosen with configutation mechanism available to each module.
        #
        # Horizontal position of the particles (lon, lat) are the
        # only properties which should not be modified directly/freely.
        # Due to the need to correct for the curvature of the coordinate
        # systems, and taking into account vector orientations, a dedicated 
        # method ``self.update_positions()`` is
        # available using x- and y-velocity as arguments:
        # E.g. to simply move particles with ambient current:
        #
        # self.update_positions(self.environment.x_sea_water_velocity,
        #                       self.environment.y_sea_water_velocity)``
        #
        # However, instead of using this method directly, most
        # OpenDrift modules use a set of ready-made and reusable
        # advection methods, added below. 

        # Advection with horizontal ocean currents
        self.advect_ocean_current()

        # Advection with vertical ocean currents
        self.vertical_advection()

        # Additional advection due to wind shear in upper 10cm of ocean:
        # ``wind_drift_factor`` multiplied by wind vector
        self.advect_wind()

        # If Stokes drift is not available from a reader,
        # it is parameterized by wind
        self.stokes_drift()

        # Vertical turbulent mixing, taking into account:
        # - element property ``terminal_velocity`` (buoyancy)
        # - environment property ``ocean_vertical_diffusivity``
        # - interaction with surface through method ``surface_wave_mixing``  (should be renamed)
        # - interaction with seafloor through method ``bottom_interaction``
        self.vertical_mixing()

    def prepare_run(self):
        """Code to be run before a simulation loop (``update()``)"""

    def bottom_interaction(self, seafloor_depth):
        """Sub method of vertical_mixing, determines settling"""



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2020, Knut-Frode Dagestad, MET Norway

from opendrift.readers.basereader import BaseReader, ContinuousReader
import numpy as np

class Reader(BaseReader, ContinuousReader):
    '''Reader providing the nearest value in time from a given time series.'''

    def __init__(self, parameter_value_map):
        '''init with a map {'time':, time array, 'variable_name': value, ...}'''

        self.times = parameter_value_map['time']
        if type(self.times) is not list:
            raise ValueError('time must be a list of datetime objects')
        del parameter_value_map['time']

        for key, var in parameter_value_map.items():
            parameter_value_map[key] = np.atleast_1d(var)
            if len(parameter_value_map[key]) != len(self.times):
                raise ValueError('All variables must have same length as time array')

        self._parameter_value_map = parameter_value_map
        self.variables = list(parameter_value_map.keys())
        self.proj4 = '+proj=latlong'
        self.xmin = -180
        self.xmax = 180
        self.ymin = -90
        self.ymax = 90
        self.start_time = self.times[0]
        self.end_time = self.times[-1]
        #self.end_time = None
        self.time_step = self.times[1] - self.times[0]
        self.name = 'reader_timeseries'

        # Run constructor of parent Reader class
        super(Reader, self).__init__()

    def get_variables(self, requestedVariables, time=None,
                      x=None, y=None, z=None):

        nearestTime, dummy1, dummy2, indxTime, dummy3, dummy4 = \
            self.nearest_time(time)

        variables = {'time': time, 'x': x, 'y': y, 'z': z}
        #variables.update(self._parameter_value_map)
        for var in requestedVariables:
            variables[var] = self._parameter_value_map[var][indxTime]*np.ones(x.shape)

        return variables




# Oil weathering interface for OpenDrift/OpenOil
# towards NOAA Oil library:
#   https://github.com/NOAA-ORR-ERD/OilLibrary
# Methods below are adapted from PyGnome:
#   https://github.com/NOAA-ORR-ERD/PyGnome

import numpy as np


def mass_transport_coeff(wind_speed):
    c_evap = 0.0025
    mass_transport_coeff = c_evap*np.power(wind_speed, 0.78)
    mass_transport_coeff[wind_speed >= 10] = \
        0.06*c_evap*np.power(wind_speed[wind_speed >= 10], 2)
    return mass_transport_coeff


def evap_decay_constant(substance, wind_speed, sea_water_temperature,
                        area, mass_components):
    K = mass_transport_coeff(wind_speed)  # per element
    f_diff = 1.0
    # vp per element, subcomponent
    vp = np.array([substance.vapor_pressure(t)
                   for t in sea_water_temperature])
    # evaporation expects mw in kg/mol, database is in g/mol
    mw = substance.molecular_weight/1000.
    # sum of mass components, per element
    sum_mi_mw = (mass_components/mw).sum(axis=1)

    gas_constant = 8.314
    decay = (-(area*f_diff*K) / (gas_constant*sea_water_temperature*
                                 sum_mi_mw)).reshape(-1, 1) * vp
    return decay


def water_uptake_coefficient(substance, wind_speed):
    # water uptake rate constant - from database
    K0Y = substance.k0y
    drop_max = 1.0e-5
    k_emul = 6.0 * K0Y * wind_speed * wind_speed / drop_max
    return k_emul



"""
Main class that represents an oil record.

This maps to the JSON used in the DB

Having a Python class makes it easier to write importing, validating etc, code.
"""
from dataclasses import dataclass, field

from ..common.utilities import dataclass_to_json, JSON_List

from ..common.measurement import VolumeFraction

from .distillation import Distillation
from .metadata import SampleMetaData
from .physical_properties import PhysicalProperties
from .environmental_behavior import EnvironmentalBehavior
from .sara import Sara
from .ests_fractions import ESTSFractions

# need the namespace for ccme, as using the fieldname the same as the class
# breaks things.
from . import ccme
from .compound import CompoundList
from .industry_property import IndustryPropertyList
from .bulk_composition import BulkCompositionList

from .validation.warnings import WARNINGS
from .validation.errors import ERRORS


@dataclass_to_json
@dataclass
class Sample:
    """
    represents the physical and chemical properties of particular sub sample.

    could be fresh oil, or weathered samples, or distillation cuts, or ...
    """
    metadata: SampleMetaData = field(default_factory=SampleMetaData)

    cut_volume: VolumeFraction = None  # from Exxon data

    physical_properties: PhysicalProperties = field(default_factory=PhysicalProperties)

    environmental_behavior: EnvironmentalBehavior = field(
        default_factory=EnvironmentalBehavior)

    SARA: Sara = field(default_factory=Sara)

    distillation_data: Distillation = field(default_factory=Distillation)

    compounds: CompoundList = field(default_factory=CompoundList)

    bulk_composition: BulkCompositionList = field(default_factory=BulkCompositionList)

    industry_properties: IndustryPropertyList = field(default_factory=IndustryPropertyList)

    headspace_analysis: CompoundList = field(default_factory=CompoundList)

    CCME: ccme.CCME = field(default_factory=ccme.CCME)

    ESTS_hydrocarbon_fractions: ESTSFractions = None

    miscellaneous: CompoundList = field(default_factory=CompoundList)

    # a place to store arbitrary extra data.
    extra_data: dict = field(default_factory=dict)

    def __post_init__(self):
        metadata = self.metadata
        if metadata.name is not None:
            if metadata.name.lower() == 'whole crude':
                metadata.name = 'Fresh Oil Sample'

        if metadata.short_name is None and metadata.name is not None:
            if metadata.name.lower() == 'fresh oil sample':
                metadata.short_name = 'Fresh Oil'
            else:
                metadata.short_name = f'{self.name[:12]}...'

    def __repr__(self):
        return f"sample: {self.metadata.name}\n{self.metadata.description}"


class SampleList(JSON_List):
    item_type = Sample

    def validate(self):
        msgs = []
        # make sure there's at least one subsample
        if len(self) == 0:
            msgs.append(ERRORS["E031"])
        else:
            # check for densities
            # note: would be good to be smart about the temp densities are at
            # this is here because only need to check the "fresh" sample
            if (self[0].physical_properties is None
                    or not self[0].physical_properties.densities):
                msgs.append(WARNINGS["W006"])

            # # check_for_distillation_cuts
            # try:
            #     if not self[0].distillation_data.cuts:
            #         msgs.append(WARNINGS['W007'])
            # except AttributeError:
            #     msgs.append(WARNINGS['W007'])

        # validate all the subsamples
        for ss in self:
            msgs.extend(ss.validate())
        return msgs



"""
Reader combined with number.
"""

from numbers import Number
from types import LambdaType
from ..basereader import BaseReader

class Combined:
    """
    A reader combined with a number.
    """

    n: Number
    r: BaseReader
    op: LambdaType

    def __init__(self, n, r, op, descop = '|'):
        self.n = n
        self.r = r
        self.op = op

        assert isinstance(n, Number)
        assert isinstance(r, BaseReader)

        self.name = f'NumCombined({self.r} {descop} {self.n})'

    @staticmethod
    def add(n, r):
        return Combined(n, r, lambda x: n + x, '+')

    @staticmethod
    def mul(n, r):
        return Combined(n, r, lambda x: n * x, '*')

    @staticmethod
    def sub(n, r):
        return Combined(n, r, lambda x: x - n, '-')

    @staticmethod
    def div(n, r):
        return Combined(n, r, lambda x: x / n, '/')


    def __getattr__(self, attr):
        """
        Forward all other method calls and attributes to reader.
        """
        return getattr(self.r, attr)

    def get_variables_interpolated(self, variables, *args, **kwargs):
        assert set(variables).issubset(self.r.variables), f"{vars} is not a subset of variables in {self.r}"

        env, env_profiles = self.r.get_variables_interpolated(variables, *args, **kwargs)

        variables = [
            var for var in env.keys() if var not in ['x', 'y', 'z', 'time']
        ]

        for var in variables:
            env[var] = self.op(env[var])

        if env_profiles is not None:
            variables = [
                var for var in env_profiles.keys() if var not in ['x', 'y', 'z', 'time']
            ]
            for var in variables:
                env_profiles[var] = self.op(env_profiles[var])

        return env, env_profiles




#!/usr/bin/env python
#
# This file is part of OpenDrift.
# 
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
# 
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
# 
# Copyright 2015, Knut-Frode Dagestad, MET Norway

import argparse
import numpy as np
import opendrift

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('filename1',
                        help='<OpenDrift filename (netCDF) of first file>')
    parser.add_argument(dest='filename2', nargs='?',
                        help='<OpenDrift filename (netCDF) of second file>')
    parser.add_argument('-b', dest='buffer',
                        default=1.0,
                        help='Buffer around plot in degrees lon/lat.')
    parser.add_argument('-f', dest='outfile',
                        default=None,
                        help='Save animation to filename.')


    args = parser.parse_args()


    o1 = opendrift.open(args.filename1)
    print(o1)

    if args.filename2 is None:
        o1.animation_profile(filename=args.outfile)
    else:
        o2 = opendrift.open(args.filename2)
        print(o2)

        # Animate and compare the two runs
        o1.animation_profile(compare=o2, filename=args.outfile,
                             legend=[args.filename1, args.filename2])



from .ops import Combine, Filter

class FilterVariables(Combine, Filter):
    """
    A reader where only some variables are forwarded.
    """

    v = None

    @property
    def variables(self):
        return self.v

    def __init__(self, r, vars):
        self.r = r

        assert set(vars).issubset(self.r.variables), f"{vars} is not a subset of variables in {self.r}"

        self.v = vars

        self.name = f'Filter({self.r} | {self.v})'

    def __getattr__(self, attr):
        """
        Forward all other method calls and attributes to reader.
        """
        return getattr(self.r, attr)



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2021, Gaute Hope, MET Norway

from importlib import resources
from pathlib import Path
import logging
import json
import itertools
from functools import lru_cache

from .oil import OpendriftOil

logger = logging.getLogger(__name__)


@lru_cache(maxsize=None)
def __get_archive__():
    import lzma

    oils = []

    with resources.open_binary('opendrift.models.openoil.adios',
                               'oils.xz') as archive:
        with lzma.open(archive, 'rt') as c:
            oils = json.load(c)

    # Add additional oils
    for f in resources.contents('opendrift.models.openoil.adios.extra_oils'):
        if Path(f).suffix == '.json':
            o = json.loads(resources.read_text('opendrift.models.openoil.adios.extra_oils', f))
            o = { 'data': { 'attributes' : o } }
            o['data']['_id'] = o['data']['attributes']['oil_id']
            o['data']['attributes']['metadata']['location'] = 'NORWAY'
            logger.debug(f"Adding additional oil: {f}..: {o['data']['_id']}, {o['data']['attributes']['metadata']['name']}")
            oils.append(o)

    return oils

def get_oil_names(location=None):
    a = __get_archive__()

    if location is not None:
        a = filter(
            lambda o: o['data']['attributes']['metadata'].get(
                'location', None) == location, a)

    return [o['data']['attributes']['metadata']['name'] for o in a]


def oils(limit=50, query=''):
    oils = filter(
        lambda o: query in o['data']['attributes']['metadata']['name'],
        __get_archive__())
    return list(OpendriftOil(o) for o in itertools.islice(oils, limit))


def find_full_oil_from_name(name) -> 'OpendriftOil':
    logger.info(f'Querying ADIOS database for oil: {name}')
    o = oils(query=name)

    if len(o) > 1:
        logger.warning(
            f"Several oils found with name: {name}: {[oo.id for oo in o]}, using first."
        )
    elif len(o) < 1:
        raise ValueError(f"No oil found with name: {name}")

    return o[0]


def get_full_oil_from_id(_id) -> 'OpendriftOil':
    logger.debug(f"Fetching full oil: {_id}")
    oils = filter(lambda o: _id == o['data']['_id'], __get_archive__())
    oil = next(OpendriftOil(o) for o in oils)
    return oil



"""
Location Coordinates Object

When saved out as JSON, this is a subset of geoJSON

i.e. a geoJSON geometry, but can only be a Point or a Polygon.

So either::

      {
        'type': 'Point',
        'coordinates': [0, 0]
      }

or::

      {
        'type': 'Polygon',
        'coordinates': [[[-5e6, -1e6], [-4e6, 1e6], [-3e6, -1e6]]]
      }

"""
from dataclasses import dataclass, field

from ..common.utilities import dataclass_to_json

@dataclass_to_json
@dataclass
class LocationCoordinates:
    type: str
    coordinates: list

    def __post_init__(self):
        '''
        Put any validation code here (__init__() is auto-generated by the
        dataclass decorator, and it will clobber any attempt to overload
        the __init__.)
        '''
        coords = self.coordinates
        if self.type == "Point":
            try:
                if len(coords) != 2:
                    raise ValueError("Point type must have two coordinates")
            except TypeError:
                raise ValueError("Point type must have two coordinates")
            try:
                self.coordinates = (float(coords[0]), float(coords[1]))
            except (ValueError, TypeError):
                raise ValueError("Point type coordinates must be a pair of numbers")
        elif self.type == "Polygon":
            try:
                ring = coords[0]
                x, y = ring[0]
            except TypeError:
                raise ValueError("Polygon coordinates must be a list of lists of points")
            if len(ring) < 4:
                raise ValueError("Polygon coordinates must be a least four points")
            if ring[0] != ring[-1]:
                raise ValueError("First and last point of a polygon must be equal")
            for p in ring:
                if len(p) != 2:
                    raise ValueError("All points of a polygon must have two coordinates")
        else:
            raise ValueError('LocationCoordinates type must be either "Point" or "Polygon"')





# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2019, Ole Baadshaug, MET Norway & Ron Saper, Carleton University Canada.
#
# Caveat: This copyright will not interfere with the open nature of OpenDrift and OpenBerg

"""
OpenBerg is an iceberg drift module bundled within the OpenDrift framework. It is a 2D- drift model, but utilizes 3D current data. The latest version of the module is an improved version of a model initially created by Ron Saper at the Carleton University as a part of a larger project funded by the MITACS Organization.

See :ref:`sphx_glr_gallery_example_openberg_det.py` for an example of a deterministic simulation.

Statistical modeling of current velocity
########################################

The reader :mod:`opendrift.readers.reader_current_from_track` is designed specifically for iceberg drift modeling. The reader uses observed positions and (if available) wind data to extrapolate the current velocity. The reader creates a uniform current field equal to the average residual speed (after subtracting wind component) of the iceberg between two observations.

This reader allows for a statistical or partly statistical modeling of iceberg drift when used with the OpenBerg module. An example script utilizing this reader can be found in :ref:`sphx_glr_gallery_example_openberg_stat.py`.

.. _openberg_parameters:

Parameters and iceberg properties affecting drift
#################################################

Icebergs are advected at a constant fraction of the wind velocity, the default setting is ``wind_drift_factor = 0.018``.

The module accounts for iceberg geometry by creating a composite iceberg using the method described by `Barker et. al. (2004) <https://www.researchgate.net/publication/44062061_Determination_of_Iceberg_Draft_Mass_and_Cross-Sectional_Areas>`_, where the geometry is described as a function of the waterline length and the keel depth of the iceberg. For further information please refer to `Barker et. al. (2004)`.

The default settings for the geometry is  ``water_line_length = 90.5`` and ``keel_depth = 60``. The composite iceberg is used to calculate a weighted average of the current velocity across the iceberg keel.

`water_line_length` is normally used to describe the length of a ship where it sits in the water. It should therefore be taken not as the circumference, but the width (or length) of the iceberg (presumably at its longest cross-section).

`keel_depth` is the depth of the ice berg from the water line. Ice bergs usually have a density of `0.92 g/mL`, sea water has a density of about `1.03 g/mL`. This means that about 90% of the ice berg mass is below the water. For a reasonably symmetric iceberg the keel depth can be estimated naively to be 9 times greater than the observed height above the water line.

The values of `wind_drift_factor`, `water_line_length` and `keel_depth` may be explicitly altered during seeding, e.g.:

.. code::

    o.seed_elements(4, 62, time=datetime.now(),
                    water_line_length=100, keel_depth=90, wind_drift_factor=0.02)


Reference: Barker, A., Sayed, M., Carrieres, T., et al. (2004). `Determination of iceberg draft, mass and cross-sectional areas <https://www.researchgate.net/publication/44062061_Determination_of_Iceberg_Draft_Mass_and_Cross-Sectional_Areas>`_.

"""


import numpy as np
import sys
from scipy.interpolate import interp1d
import pyproj
import logging; logger = logging.getLogger(__name__)

from opendrift.models.basemodel import OpenDriftSimulation
from opendrift.elements.elements import LagrangianArray
from opendrift.readers.basereader import BaseReader

# Defining the iceberg element properties
class IcebergObj(LagrangianArray):
    """Extending LagrangianArray with variables relevant for iceberg objects.

    """

    # We add the properties to the element class
    variables = LagrangianArray.add_variables([
        ('wind_drift_factor', {'dtype': np.float32,
                               'units': '1',
            'description': 'The fraction of the wind speed at which an iceberg is moved',
                       	       'default': 0.018}),
        ('water_line_length', {'dtype': np.float32,				# Iceberg size
                               'units': 'm',
                               'default': 90.5}),
        ('keel_depth', {'dtype': np.float32,					# Iceberg keel depth
                        'units': 'm',
                        'default': 60.})])

class OpenBerg(OpenDriftSimulation):
    """The Deterministic iceberg model in the OpenDrift framework.

        Advects an iceberg with the current at all available depths and
        as a function of the wind vector according to the above and below
        water cross-sectional profile of the object.
    """

    ElementType = IcebergObj

    required_variables = {
        'x_wind': {'fallback': 0},
        'y_wind': {'fallback': 0},
        'x_sea_water_velocity': {'fallback': 0, 'profiles': True},
        'y_sea_water_velocity': {'fallback': 0, 'profiles': True},
        'land_binary_mask': {'fallback': None},
        }

    # Default colors for plotting
    status_colors = {'initial': 'green', 'active': 'blue',
                     'missing_data': 'gray', 'stranded': 'red'}

    # Configuration
    def __init__(self, d=None, label=None, *args, **kwargs):
        self.name = 'OpenBerg'
        self.label=label

        #self.required_profiles = ['x_sea_water_velocity',
        #                          'y_sea_water_velocity']  # Get vertical current profiles

        self.required_profiles_z_range = [-120, 0] # [min_depth, max_depth]

        # Calling general constructor of parent class
        super(OpenBerg, self).__init__(*args, **kwargs)

        self._add_config({
            'seed:wind_drift_factor': {'type': 'float', 'min': 0, 'max': 1,
                'default': 0.018, 'units': 'fraction',
                'level': self.CONFIG_LEVEL_ADVANCED,
                'description': 'Icebergs are moved with this fraction of the wind speed, in addition to ocean current forcing'},
            'seed:water_line_length': {'type': 'float', 'min': 0.1, 'max': 99999,
                'default': 90.5, 'units': 'meters',
                'level': self.CONFIG_LEVEL_ADVANCED,
                'description': 'Length of iceberg.'},
            'seed:keel_depth': {'type': 'float', 'min': 0.1, 'max': 1000,
                'default': 60, 'units': 'meters',
                'level': self.CONFIG_LEVEL_ADVANCED,
                'description': 'Length of iceberg keel (part belor sea surface).'},
            })

        self._set_config_default('drift:current_uncertainty', .15)
        self._set_config_default('drift:wind_uncertainty', .5)

    def seed_elements(self, *args, **kwargs):
        for var in ['wind_drift_factor', 'water_line_length', 'keel_depth']:
            if var not in kwargs:
                kwargs[var] = self.get_config('seed:' + var)

        super(OpenBerg, self).seed_elements(*args, **kwargs)

    def update(self):
        """Update positions and properties of icebergs."""
		# Move icebergs at given % of wind speed
        self.advect_wind()

        # Move icebergs as per Anne Barker et al, 2004, with weighted average of current
        # down to draft of berg.
        # For current 10n meters down, where n is the depth index from 0 to 19

        # RETRIEVE CURRENT SPEED:
        y_sea_water_vel = self.environment_profiles['y_sea_water_velocity']
        x_sea_water_vel = self.environment_profiles['x_sea_water_velocity']

		# FIND THE WEIGHTED AVERAGE CURRENT SPEED ACROSS THE ICEBERG KEEL
        net_x_swv = np.zeros(x_sea_water_vel.shape[1])
        net_y_swv = np.zeros(y_sea_water_vel.shape[1])
        for indx in range(len(self.uw_weighting)):
            net_y_swv = net_y_swv +y_sea_water_vel[indx,:]*self.uw_weighting[indx]
            net_x_swv = net_x_swv +x_sea_water_vel[indx,:]*self.uw_weighting[indx]

        self.update_positions(net_x_swv,net_y_swv)


    def prepare_run(self):
        """	Model spesific preparations.
         	Set the weighting for modelled current depths as per Table 5 of Barker 2004,
        	'Determination of Iceberg Draft, Mass and Cross-Sectional Areas',
        	Proceedings of The Fourteenth (2004) International Offshore and
        	Polar Engineering Conference.

        	NB! This version of OpenBerg does not allow for seeding of iceberg elements
        	of different sizes.

        	Also controles that the model handles readers without block data correctly.
        """
        # Retrieve profile provided in z dimension by reader
        variable_groups, reader_groups, missing_variables = \
         	self.get_reader_groups(['x_sea_water_velocity','y_sea_water_velocity'])

        if len(reader_groups) == 0:
        	# No current data -> fallback values used
        	self.uw_weighting = np.array([1])
        	return

		# Obtain depth levels from reader:
        reader_name = reader_groups[0][0]
        if hasattr(self.readers[reader_name], 'z'):
            profile = np.abs(np.ma.filled(self.readers[reader_name].z))
        else:  # ROMS sigma levels
            profile = np.abs(np.ma.filled(self.readers[reader_name].zlevels))

        # If current data is missing in at least one dimension, no weighting is performed:
        if len(missing_variables) > 0:
        	logger.warning('Missing current data, weigthing array set to [1]')
        	self.uw_weighting = np.array([1])
        	return

        # No need to create weighting array if only one z-level is provided:
        if len(profile) == 1:
        	self.uw_weighting = np.array([1])
        	return

        # Make copies to prevent outside value to be modified
        water_line_length = self.elements_scheduled.water_line_length.copy()
        depth = self.elements_scheduled.keel_depth.copy()

        # Calculate the weighting array corresponding to the iceberg profile
        uw_weighting = self.composite_iceberg(water_line_length=water_line_length, depth=depth)



        #### Interpolate weighting array to match z-dimension of current reader ###

        # Array of "Barker-depths" (10n)m, , where n is the depth index from 0 to 19
        x = np.linspace(0,(len(uw_weighting)-1)*10,len(uw_weighting))

        # Create a linear interpolator
        interpol = interp1d(x,uw_weighting)

        # Obtain corresponding depth levels from reader:
        self.reader_z_profile = profile[(profile >= 0) & (profile <= x.max())]

        if len(profile[profile < 0]) > 0:
        	logger.warning('Current reader containing currents above water!'
        						'Weighting of current profile may not work!')

        # Interpolate and normalize weights:

        ###### NB!  Interpolator only includes values from reader within the range of the
        ######		Barker-depth array. Meaning values just outside is not used for interpolation.
        ######		Ex.: If x.max=195 and the current reader includes data for
        ######		the z-profile: [0,3,10,15,25,50,75,100,150,200] only data from the
        ######		z-levels [0,3,10,15,25,50,75,100,150] are used.

        interpol_weight = interpol(self.reader_z_profile)
        normalized_weight = interpol_weight/interpol_weight.sum()

        if normalized_weight.all():
            self.uw_weighting = normalized_weight

        super(OpenBerg, self).prepare_run()

    def composite_iceberg(self, water_line_length=90.5, depth=60):

    	"""This function creates a weigthing array for the current across the keel of an
	    	iceberg based on waterline length and keel depth. The function uses the parameters
    		in table 5 from Barker et. al.(2004).
    	"""
		# Parameters from Baker et. al.:

    	a_param = [9.5173,11.1717,12.4798,13.6010,14.3249,13.7432,13.4527,15.7579,
    				14.7259,11.8195,11.3610,10.9202,10.4966,10.0893,9.6979,9.3216,8.9600,
    				8.6124,8.2783,7.9571]

    	b_param = [-25.94,-107.50,-232.01,-344.60,-456.57,-433.33,-519.56,-1111.57,-1125.00,
    					-852.90,-931.48,-1007.02,-1079.62,-1149.41,-1216.49,-1280.97,
    					-1342.95,-1402.52,-1459.78,-1514.82]

    	d = int(round(depth/10))

    	if d < 0: #Incase depth is given as a negative value
    		d = -d


    	if d > len(a_param):
    		d = len(a_param)
    		print('##### OpenBerg does not support icebergs with keel depths greater than 200m!\n' +
				'Using a composite iceberg with given waterline length and keel depth 200m')

    	area_list=[]

    	for i in range(0,d):
    		A = self.lin_func(a_param[i],b_param[i],water_line_length)
    		area_list.append(A)

    	A_list = np.array(area_list)

    	# Normalize array such that it sums to 1:
    	weigthing_array = np.array(area_list)/sum(np.array(area_list))

    	return weigthing_array


    def lin_func(self,a,b,L):
    	"""Returns value of linear function A=aL+b."""
    	A = a*L + b

    	return A



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2020, Gaute Hope, MET Norway
# Copyright 2020, Simon Weppe, MetOcean Solution, MetService New Zealand
# Copyright 2015, Knut-Frode Dagestad, MET Norway

import numpy as np
from datetime import datetime, timedelta, timezone
from netCDF4 import Dataset, MFDataset, num2date
from scipy.interpolate import LinearNDInterpolator
import scipy, scipy.spatial
import pyproj
import logging
logger = logging.getLogger(__name__)

from opendrift.readers.basereader import BaseReader, UnstructuredReader


class Reader(BaseReader, UnstructuredReader):
    """
    A reader for unstructured (irregularily gridded) `CF compliant
    <https://cfconventions.org/>`_ netCDF files.

    Args:
        :param filename: A single netCDF file, or a pattern of files. The
                         netCDF file can also be an URL to an OPeNDAP server.
        :type filename: string, requiered.

        :param name: Name of reader
        :type name: string, optional

        :param proj4: PROJ.4 string describing projection of data.
        :type proj4: string, optional

    .. seealso::

        py:mod:`opendrift.readers.basereader.unstructured`.
    """

    # Misspelled standard names in some (Akvaplan-NIVA) FVCOM files
    variable_aliases = {
        'eastward_sea_water_velocity': 'x_sea_water_velocity',
        'Northward_sea_water_velocity': 'y_sea_water_velocity',
        'eastward wind': 'x_wind',
        'northward wind': 'y_wind',
        'ww': 'upward_sea_water_velocity',
    }

    node_variables = [
        'salinity',
        'temperature',
        'sea_floor_depth_below_sea_level',
        'sea_floor_depth_below_geoid',
        'sea_surface_height_above_geoid',
    ]

    face_variables = [
        'x_sea_water_velocity',
        'y_sea_water_velocity',
        'upward_sea_water_velocity',
    ]

    dataset = None

    # For in-memory caching of Sigma-coordinates and ocean depth
    siglay = None
    siglev = None
    siglay_center = None
    siglev_center = None
    ocean_depth_nele = None
    ocean_depth_node = None

    def __init__(self, filename=None, name=None, proj4=None):
        if filename is None:
            raise ValueError('Filename is missing')
        filestr = str(filename)
        if name is None:
            self.name = filestr
        else:
            self.name = name

        # xarray currently does not handle this type of grid:
        # https://github.com/pydata/xarray/issues/2233

        self.timer_start("open dataset")
        logger.info('Opening dataset: ' + filestr)
        if ('*' in filestr) or ('?' in filestr) or ('[' in filestr):
            logger.info('Opening files with MFDataset')
            self.dataset = MFDataset(filename)
        else:
            logger.info('Opening file with Dataset')
            self.dataset = Dataset(filename, 'r')

        if proj4 is not None:
            logger.info('Using custom projection: %s..' % proj4)
            self.proj4 = proj4
        else:
            self.proj4 = self.dataset.CoordinateProjection.strip()
            if self.proj4 == 'none': self.proj4 = None
            logger.info('Reading projection..: %s', self.proj4)
            assert self.proj4 is not None and len(
                self.proj4
            ) > 0, "No projection in data-file, please supply to reader"

        logger.info('Reading grid and coordinate variables..')
        assert self.dataset.CoordinateSystem == "Cartesian", "Only cartesian coordinate systems supported"

        self.x = self.dataset['x'][:]
        self.y = self.dataset['y'][:]
        self.xc = self.dataset['xc'][:]
        self.yc = self.dataset['yc'][:]

        # Times in FVCOM files are 'Modified Julian Day (MJD)', or fractional days since
        # 1858-11-17 00:00:00 UTC
        assert self.dataset['time'].time_zone == 'UTC'
        assert self.dataset['time'].units == 'days since 1858-11-17 00:00:00'
        assert self.dataset['time'].format == 'modified julian day (MJD)'
        ref_time = datetime(1858, 11, 17, 00, 00,
                            00)  # TODO: include , tzinfo=timezone.utc)
        self.times = np.array([
            ref_time + timedelta(days=d.item())
            for d in self.dataset['time'][:]
        ])
        self.start_time = self.times[0]
        self.end_time = self.times[-1]
        # time steps are not constant

        self.xmin = np.min(self.x)
        self.xmax = np.max(self.x)
        self.ymin = np.min(self.y)
        self.ymax = np.max(self.y)

        self.variable_mapping = {}
        for var_name in self.dataset.variables:
            # skipping coordinate variables
            if var_name in [
                    'x', 'y', 'time', 'lon', 'lat', 'lonc', 'latc', 'siglay',
                    'siglev', 'siglay_center', 'siglev_center'
            ]:
                continue

            # We only use sea_floor at nodes
            if var_name in ['h_center']:
                continue

            var = self.dataset[var_name]
            if 'standard_name' in var.ncattrs():
                std_name = var.getncattr('standard_name')
                std_name = self.variable_aliases.get(std_name, std_name)
                self.variable_mapping[std_name] = str(var_name)

        self.variables = list(self.variable_mapping.keys())

        # Run constructor of parent Reader class
        super().__init__()

        self.boundary = self._build_boundary_polygon_(self.x.compressed(),
                                                      self.y.compressed())

        self.timer_start("build index")
        logger.debug("building index of nodes..")
        self.nodes_idx = self._build_ckdtree_(self.x, self.y)

        logger.debug("building index of faces..")
        self.faces_idx = self._build_ckdtree_(self.xc, self.yc)

        self.timer_end("build index")

        self.timer_end("open dataset")

    def plot_mesh(self):
        """
        Plot the grid mesh. Does not automatically show the figure.
        """
        import matplotlib.pyplot as plt
        plt.figure()
        plt.scatter(self.x, self.y, marker='x', color='blue', label='nodes')
        plt.scatter(self.xc,
                    self.yc,
                    marker='o',
                    color='red',
                    label='centroids')

        x, y = getattr(self.boundary, 'context').exterior.xy
        plt.plot(x, y, color='green', label='boundary')

        plt.legend()
        plt.title('Unstructured grid: %s\n%s' % (self.name, self.proj))
        plt.xlabel('x [m]')
        plt.ylabel('y [m]')

    def get_variables(self,
                      requested_variables,
                      time=None,
                      x=None,
                      y=None,
                      z=None):
        """
        FVCOM Grid:

        FVCOM uses 'triangular prisms' for gridding. Some variables are defined
        on the faces of the triangles, while others at the node.

        `x` and `y` holds the positions of the node, while `xc` and `yc` holds
        the positions on the centroids/faces. The centroids/faces are also
        known as 'zonal', or elements (presumably as in finite element).

        .. note::

            Currently this reader does not really interpolate. It looks up the
            closest point in time and space.

        Each element has a lookup-table of its surrounding elements, this list can be
        used when looking up elements for the interpolator of an arbitrary
        point on the grid. The same goes for the nodes.

        Let E be number of elements and N be number of nodes.

        Relevant lookup-tables:

        nbe:        [3 x E]  elements surround each element
        nbve:       [9 x N]  elements surrounding each node, minimum 3
        nbsn:       [11 x N] nodes surrounding each node

        Variables:

        Face:
        * u
        * v

        Node:
        * temperature
        * salinity

        """

        x = np.atleast_1d(x)
        y = np.atleast_1d(y)
        z = np.atleast_1d(z)
        if len(z) == 1:
            z = z[0] * np.ones(x.shape)

        logger.debug("Requested variabels: %s, lengths: %d, %d, %d" %
                     (requested_variables, len(x), len(y), len(z)))

        requested_variables, time, x, y, z, _outside = \
            self.check_arguments(requested_variables, time, x, y, z)

        nearest_time, _time_before, _time_after, indx_nearest, _indx_before, _indx_after = self.nearest_time(
            time)

        logger.debug("Nearest time: %s" % nearest_time)

        node_variables = [
            var for var in requested_variables if var in self.node_variables
        ]
        face_variables = [
            var for var in requested_variables if var in self.face_variables
        ]

        assert (len(node_variables) + len(face_variables)
                ) == len(requested_variables), "missing variables requested"

        variables = {}

        if node_variables:
            logger.debug("Interpolating node-variables..")

            nodes = self._nearest_node_(x, y)
            assert len(nodes) == len(x)

            for var in node_variables:
                dvar = self.variable_mapping.get(var)
                logger.debug("Interpolating: %s (%s)" % (var, dvar))
                dvar = self.dataset[dvar]

                # sigma ind depends on whether variable is defined on sigma layer og sigma level
                if 'siglev' in dvar.dimensions:
                    sigma_ind = self.__nearest_node_sigma__(dvar, nodes, z)
                    assert len(sigma_ind) == len(z)

                    # Reading the smallest block covering the actual data
                    block = dvar[indx_nearest,
                                 slice(sigma_ind.min(),
                                       sigma_ind.max() + 1),
                                 slice(nodes.min(),
                                       nodes.max() + 1)]
                    # Picking the nearest value
                    variables[var] = block[sigma_ind - sigma_ind.min(),
                                           nodes - nodes.min()]
                else:  # no depth dimension
                    variables[var] = dvar[slice(nodes.min(), nodes.max() + 1)]

        if face_variables:
            logger.debug("Interpolating face-variables..")

            fcs = self._nearest_face_(x, y)
            assert len(fcs) == len(x)

            for var in face_variables:
                dvar = self.variable_mapping.get(var)
                logger.debug("Interpolating: %s (%s)" % (var, dvar))
                dvar = self.dataset[dvar]

                # sigma ind depends on whether variable is defined on sigma layer og sigma level
                if 'siglay' in dvar.dimensions:
                    sigma_ind = self.__nearest_face_sigma__(dvar, fcs, z)
                    assert len(sigma_ind) == len(z)

                    # Reading the smallest profile slice covering the actual data
                    block = dvar[indx_nearest,
                                 slice(sigma_ind.min(),
                                       sigma_ind.max() + 1),
                                 slice(fcs.min(),
                                       fcs.max() + 1)]
                    # Picking the nearest value
                    variables[var] = block[sigma_ind - sigma_ind.min(),
                                           fcs - fcs.min()]
                else:  # no depth dimension
                    variables[var] = dvar[slice(fcs.min(), fcs.max() + 1)]

        return variables

    @staticmethod
    def _vector_nearest_(X, xp):
        """
        Find nearest element in vector of vectors `X` for each `xp`.

        Args:

            X   NxM matrix of levels
            xp  M   vector of positions

        Returns:
            i   M   vector of indices [0..N] of closest element in
                    X[0..N, i] to xp[i]
        """
        xp = np.atleast_2d(xp)
        return np.argmin(np.abs(X - xp), axis=0)

    def __nearest_node_sigma__(self, var, nodes, z):
        """
        Find nearest depth at node (sigma layer or sigma level depending on where the variable is defined).
        """
        shp = var.shape

        if self.siglay is None:
            logger.debug('Reading siglays into memory...')
            self.siglay = self.dataset['siglay'][:]

        if self.siglev is None:
            logger.debug('Reading siglevs into memory...')
            self.siglev = self.dataset['siglev'][:]

        if shp[1] == self.siglay.shape[0]:
            sigmas = self.siglay[:, nodes]
        else:
            sigmas = self.siglev[:, nodes]

        if self.ocean_depth_node is None:
            logger.debug('Reading ocean depth into memory...')
            self.ocean_depth_node = self.dataset['h'][:]

        # Calculating depths from sigmas
        depths = self.z_from_sigma(sigmas, self.ocean_depth_node[nodes])

        return self._vector_nearest_(depths, z)

    def __nearest_face_sigma__(self, var, el, z):
        """
        Find nearest depth at element (sigma layer or sigma level depending on where the variable is defined).
        """
        shp = var.shape

        if self.siglay_center is None:
            logger.info('Reading siglay_centers into memory...')
            self.siglay_center = self.dataset['siglay_center'][:]

        if self.siglev_center is None:
            logger.info('Reading siglev_centers into memory...')
            self.siglev_center = self.dataset['siglev_center'][:]

        if shp[1] == self.siglay_center.shape[0]:
            sigmas = self.siglay_center[:, el]
        else:
            sigmas = self.siglev_center[:, el]

        if self.ocean_depth_nele is None:
            logger.info('Reading ocean depth center into memory...')
            self.ocean_depth_nele = self.dataset['h_center'][:]

        # Calculating depths from sigmas
        depths = self.z_from_sigma(sigmas, self.ocean_depth_nele[el])

        return self._vector_nearest_(depths, z)

    @staticmethod
    def z_from_sigma(sigma, depth, elevation=0):
        """
        Calculate z-depth from sigma constant.

        https://rdrr.io/github/edwardlavender/fvcom.tbx/src/R/depth_from_known.R

        Args:

            sigma       Sigma coefficient(s)
            depth       Depth below mean sea-surface
            elevation   Elevation of sea-surface (e.g. tidal)

        Returns: z, depth below sea-surface in meters.
        """
        return sigma * (depth + elevation)




import logging
import numpy as np
from scipy.ndimage import map_coordinates, grey_dilation
import logging; logging.captureWarnings(True); logger = logging.getLogger(__name__)
from scipy.interpolate import interp1d, LinearNDInterpolator

logger = logging.getLogger('opendrift')  # using common logger

def expand_numpy_array(data):
    if isinstance(data, np.ma.MaskedArray):
        logger.warning('Converting masked array to numpy array before interpolating')
        data = np.ma.filled(data, fill_value=np.nan)
    if not np.isfinite(data).any():
        logger.warning('Only NaNs, returning')
        return
    mask = ~np.isfinite(data)
    data[mask] = np.finfo(np.float64).min
    data[mask] = grey_dilation(data, size=3)[mask]
    data[data==np.finfo(np.float64).min] = np.nan


###########################
# 2D interpolator classes
###########################

class Nearest2DInterpolator():

    def __init__(self, xgrid, ygrid, x, y):
        self.x = x
        self.y = y
        self.xi = (x - xgrid.min())/(xgrid.max()-xgrid.min())*len(xgrid)
        self.yi = (y - ygrid.min())/(ygrid.max()-ygrid.min())*len(ygrid)
        self.xi = np.round(self.xi).astype(np.uint32)
        self.yi = np.round(self.yi).astype(np.uint32)
        self.xi[self.xi >= len(xgrid)] = len(xgrid)-1
        self.yi[self.yi >= len(ygrid)] = len(ygrid)-1

    def __call__(self, array2d):
        return array2d[self.yi, self.xi]


class NDImage2DInterpolator():

    def __init__(self, xgrid, ygrid, x, y):
        self.x = x
        self.y = y
        self.xi = (x - xgrid.min())/(xgrid.max()-xgrid.min())*len(xgrid)
        self.yi = (y - ygrid.min())/(ygrid.max()-ygrid.min())*len(ygrid)

    def __call__(self, array2d):
        try:
            array2d = np.ma.array(array2d, mask=array2d.mask)
            array2d[array2d.mask] = np.nan  # Gives holes
        except:
            pass
        return np.ma.masked_invalid(
            map_coordinates(array2d, [self.yi, self.xi],
                            cval=np.nan, order=0))


class LinearND2DInterpolator():


    logger = logging.getLogger('opendrift')

    def __init__(self, xgrid, ygrid, x, y):
        self.block_x, self.block_y = np.meshgrid(xgrid, ygrid)
        self.block_x = self.block_x.ravel()
        self.block_y = self.block_y.ravel()
        self.x = x
        self.y = y

    def __call__(self, array2d):
        array_ravel = array2d.ravel()
        valid = np.isfinite(array_ravel)
        #if isinstance(array2d.mask, np.ndarray):
        #    valid = ~array2d.ravel().mask
        #elif array2d.mask == False:
        #    valid = np.ones(array_ravel.shape, dtype=bool)
        #elif array2d.mask == True:
        #    valid = np.zeros(array_ravel.shape, dtype=bool)

        if hasattr(self, 'interpolator'):
            if not np.array_equal(valid, self.interpolator.valid):
                logger.debug('Cannot reuse interpolator - validity of '
                              'array is different from original.')
        if hasattr(self, 'interpolator') and (np.array_equal(
                 valid, self.interpolator.valid)):
            # Reuse stored interpolator with new data
            self.interpolator.values[:, 0] = \
                (array_ravel[valid])
        else:
            # Make new interpolator for given x,y
            self.interpolator = LinearNDInterpolator(
                (self.block_y[valid],
                 self.block_x[valid]),
                array_ravel[valid])
            # Store valid array, to determine if can be used again
            self.interpolator.valid = valid
            # Call interpolator to avoid threading-problem:
            # https://github.com/scipy/scipy/issues/8856
            self.interpolator((0,0))

        return self.interpolator(self.y, self.x)


class Linear2DInterpolator():

    logger = logging.getLogger('opendrift')

    def __init__(self, xgrid, ygrid, x, y):
        self.x = x
        self.y = y
        self.xi = (x - xgrid[0])/(xgrid[-1]-xgrid[0])*(len(xgrid)-1)
        self.yi = (y - ygrid[0])/(ygrid[-1]-ygrid[0])*(len(ygrid)-1)

    def __call__(self, array2d):
        if isinstance(array2d,np.ma.MaskedArray):
            logger.debug('Converting masked array to numpy array for interpolation')
            array2d = np.ma.filled(array2d, fill_value=np.nan)
        if not np.isfinite(array2d).any():
            logger.warning('Only NaNs input to linearNDFast - returning')
            return np.nan*np.ones(len(self.xi))

        # Fill NaN-values with nearby real values
        interp = map_coordinates(array2d, [self.yi, self.xi],
                                 cval=np.nan, order=1)
        missing = np.where(~np.isfinite(interp))[0]
        i=0
        while len(missing) > 0:
            i += 1
            if i > 10:
                logger.warning('Still NaN-values after 10 iterations, exiting!')
                return interp
            logger.debug('NaN values for %i elements, expanding data %i' %
                          (len(missing), i))
            expand_numpy_array(array2d)
            interp[missing] = map_coordinates(
                array2d, [self.yi[missing], self.xi[missing]],
                cval=np.nan, order=1, mode='nearest')
            missing = np.where(~np.isfinite(interp))[0]

        return interp


horizontal_interpolation_methods = {
    'nearest': Nearest2DInterpolator,
    'ndimage': NDImage2DInterpolator,
    'linearND': LinearND2DInterpolator,
    'linearNDFast': Linear2DInterpolator}


###########################
# 1D interpolator classes
###########################

class Nearest1DInterpolator():

    def __init__(self, zgrid, z):

        # Truncating above and below
        z[z < zgrid.min()] = zgrid.min()
        z[z > zgrid.max()] = zgrid.max()
        # Interpolator zgrid -> index
        if zgrid[1] > zgrid[0]:  # increasing
            z_interpolator = interp1d(zgrid, range(len(zgrid)))
        else:  # decreasing values, must flip for interpolator
            z_interpolator = interp1d(zgrid[::-1], range(len(zgrid))[::-1])
        # Indices corresponding to nearest value in zgrid
        self.zi = np.round(z_interpolator(z)).astype(np.uint8)
        self.zi[self.zi < 0] = 0
        self.zi[self.zi >= len(zgrid)] = len(zgrid) - 1

    def __call__(self, array2d):
        return array2d[self.zi, range(len(self.zi))]


class Linear1DInterpolator():

    def __init__(self, zgrid, z):

        # Truncating above and below
        z[z < zgrid.min()] = zgrid.min()
        z[z > zgrid.max()] = zgrid.max()
        # Interpolator zgrid -> index
        if zgrid[1] > zgrid[0]:  # increasing
            z_interpolator = interp1d(zgrid, range(len(zgrid)))
        else:  # decreasing values, must flip for interpolator
            z_interpolator = interp1d(zgrid[::-1], range(len(zgrid))[::-1])
        z_interpolator(z[0])  # to prevent threading issues
        # Indices corresponding to layers above and below
        interp_zi = z_interpolator(z)
        self.index_above = np.floor(interp_zi).astype(np.int8)
        self.index_above[self.index_above < 0] = 0
        self.index_below = np.minimum(self.index_above + 1, len(zgrid) - 1)
        self.weight_above = 1 - (interp_zi - self.index_above)
        self.xi = range(len(z))

    def __call__(self, array2d):
        return array2d[self.index_above, self.xi]*self.weight_above + \
               array2d[self.index_below, self.xi]*(1 - self.weight_above)

vertical_interpolation_methods = {
    'nearest': Nearest1DInterpolator,
    'linear': Linear1DInterpolator}


def fill_NaN_towards_seafloor(array):
    """Extrapolate NaN-values (missing) towards seafloor"""
    filled = False
    for i in range(1, array.shape[0]):
        mask = np.isnan(array[i,:,:])
        if np.sum(mask) > 0:
            array[i, mask] = array[i-1, mask]
            filled = True
    return filled




# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway

import logging; logger = logging.getLogger(__name__)
from opendrift.models.basemodel import OpenDriftSimulation
from opendrift.elements.passivetracer import PassiveTracer


class WindBlow(OpenDriftSimulation):
    """Demonstration trajectory model based on OpenDrift framework.

    Simply advects a particle (passive tracer with
    no properties except for position) with the ambient wind.
    """

    ElementType = PassiveTracer
    required_variables = {
        'x_wind': {'fallback': 0},
        'y_wind': {'fallback': 0}
        }

    max_speed = 12  # m/s

    def update(self):

        # Simply move particles with ambient wind
        self.update_positions(self.environment.x_wind,
                              self.environment.y_wind)



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway

"""
Leeway is the search and rescue (SAR) model developed by the US Coast Guard, as originally described in

    Allen, A.A, 2005: Leeway Divergence, USCG R&D Center Technical Report CG-D-05-05. Available through http://www.ntis.gov, reference ADA435435 

    Allen A.A. and J.V. Plourde (1999) Review of Leeway; Field Experiments and Implementation, USCG R&D Center Technical Report CG-D-08-99. Available through http://www.ntis.gov, reference ADA366414

and later extended and modified by e.g.

    Breivik, O., A. Allen, C. Maisondieu, J.-C. Roth, and B. Forest, 2012: The leeway of shipping containers at different immersion levels. Ocean Dyn., 62, 741–752, doi:10.1007/s10236-012-0522-z

The Leeway model is based on empirically determined coefficients as tabulated in https://github.com/OpenDrift/opendrift/blob/master/opendrift/models/OBJECTPROP.DAT

The Leeway model is been reprogrammed in Python for OpenDrift by Knut-Frode Dagestad of the Norwegian Meteorological Institute.
"""

from builtins import range
import os
from collections import OrderedDict
import logging; logger = logging.getLogger(__name__)

import numpy as np

from opendrift.models.basemodel import OpenDriftSimulation
from opendrift.elements import LagrangianArray

RIGHT = 0
LEFT = 1


# Defining the leeway element properties
class LeewayObj(LagrangianArray):
    """Extending LagrangianArray with variables relevant for leeway objects.

    """
    variables = LagrangianArray.add_variables([
        ('object_type', {'dtype': np.uint16,
                        'units': '1',
                        'seed': False,
                        'default': 0}),
        ('orientation', {'dtype': np.uint8,
                         'units': '1',
            'description': '0/1 is left/right of downwind. Randomly chosen at seed time',
                         'seed': False,
                         'default': 1}),
        ('jibe_probability', {'dtype': np.float32,
                             'units': '1/h',
            'description': 'Probability per hour that an object may change orientation (jibing)',
                             'default': 0.04}),
        ('downwind_slope', {'dtype': np.float32,
                           'units': '%',
                           'seed': False,
                           'default': 1}),
        ('crosswind_slope', {'dtype': np.float32,
                            'units': '1',
                           'seed': False,
                            'default': 1}),
        ('downwind_offset', {'dtype': np.float32,
                            'units': 'cm/s',
                           'seed': False,
                            'default': 0}),
        ('crosswind_offset', {'dtype': np.float32,
                             'units': 'cm/s',
                           'seed': False,
                             'default': 0}),
        ('downwind_eps', {'dtype': np.float32,
                         'units': 'cm/s',
                           'seed': False,
                         'default': 0}),
        ('crosswind_eps', {'dtype': np.float32,
                          'units': 'cm/s',
                           'seed': False,
                          'default': 0}),
        ('current_drift_factor', {'dtype': np.float32,
                                  'units': '1',
                'description': 'Elements are moved with this fraction of the '
                            'current vector, in addition to currents '
                            'and Stokes drift',
                               'default': 1})
        ])


class Leeway(OpenDriftSimulation):
    """The Leeway model in the OpenDrift framework.

        Advects a particle (a drifting object) with the ambient current and
        as a function of the wind vector according to the drift properties
        of the object.
    """

    ElementType = LeewayObj

    required_variables = {
        'x_wind': {'fallback': None},
        'y_wind': {'fallback': None},
        'x_sea_water_velocity': {'fallback': None},
        'y_sea_water_velocity': {'fallback': None},
        'land_binary_mask': {'fallback': None},
        }


    # Default colors for plotting
    status_colors = {'initial': 'green', 'active': 'blue',
                     'missing_data': 'gray', 'stranded': 'red',
                     'evaporated': 'yellow', 'dispersed': 'magenta'}

    max_speed = 5  # Assumed max average speed of any element

    # Configuration
    def __init__(self, d=None, *args, **kwargs):

        # Read leeway object properties from file
        if d is None:
            d = os.path.dirname(os.path.realpath(__file__))
            objprop_file = open(d + '/OBJECTPROP.DAT', 'r')
        else:
            objprop_file = open(d, 'r')

        # objprop_file=open('/home/oyvindb/Pro/opendrift/OBJECTPROP.DAT','r')
        objproptxt = objprop_file.readlines()
        objprop_file.close()
        self.leewayprop = OrderedDict({})
        for i in range(len(objproptxt)//3+1):
            # Stop at first blank line
            if not objproptxt[i*3].strip():
                break
            elems = objproptxt[i*3].split()[0]

            objKey = objproptxt[i*3].split()[0].strip()
            arr = [float(x) for x in objproptxt[i*3+2].split()]
            props = {'OBJKEY': objKey}
            props['Description'] = objproptxt[i*3+1].strip()
            props['DWSLOPE'] = arr[0]
            props['DWOFFSET'] = arr[1]
            props['DWSTD'] = arr[2]
            props['CWRSLOPE'] = arr[3]
            props['CWROFFSET'] = arr[4]
            props['CWRSTD'] = arr[5]
            props['CWLSLOPE'] = arr[6]
            props['CWLOFFSET'] = arr[7]
            props['CWLSTD'] = arr[8]
            self.leewayprop[i+1] = props

        # Config
        descriptions = [self.leewayprop[p]['Description'] for
                        p in self.leewayprop]

        # Calling general constructor of parent class
        super(Leeway, self).__init__(*args, **kwargs)

        self._add_config({
            'seed:object_type': {'type': 'enum', 'enum': descriptions,
                'default': descriptions[0],
                'description': 'Leeway object category for this simulation',
                'level': self.CONFIG_LEVEL_ESSENTIAL},
            'seed:jibe_probability': {'type': 'float',
                'default': 0.04, 'min': 0, 'max': 1,
                'description': 'Probability per hour for jibing (objects changing orientation)',
                'units': 'probability', 'level': self.CONFIG_LEVEL_BASIC},
            })

        self._set_config_default('general:time_step_minutes', 10)
        self._set_config_default('general:time_step_output_minutes', 60)

    def seed_elements(self, lon, lat, object_type=None, **kwargs):
        """Seed particles in a cone-shaped area over a time period."""
        # All particles carry their own object_type (number),
        # but so far we only use one for each sim
        # objtype = np.ones(number)*object_type

        if 'number' in kwargs and kwargs['number'] is not None:
            number = kwargs['number']
        else:
            number = self.get_config('seed:number')

        if object_type is None:
            object_name = self.get_config('seed:object_type')
            # Get number from name
            found = False
            for object_type in range(1, len(self.leewayprop)+1):
                if self.leewayprop[object_type]['OBJKEY'] == object_name or(
                    self.leewayprop[object_type]['Description'] == object_name):
                    found = True
                    break
            if found is False:
                logger.info(self.list_configspec())
                raise ValueError('Object %s not available' % object_type)

        logger.info('Seeding elements of object type %i: %s (%s)' %
                     (object_type, self.leewayprop[object_type]['OBJKEY'],
                      self.leewayprop[object_type]['Description']))

        # Drift orientation of particles.  0 is right of downwind,
        # 1 is left of downwind
        # orientation = 1*(np.random.rand(number)>=0.5)
        # Odd numbered particles are left-drifting, even are right of downwind.
        orientation = np.r_[:number] % 2
        ones = np.ones_like(orientation)

        # Downwind leeway properties.
        # Generate normal, N(0,1), random perturbations for leeway coeffs.
        # Negative downwind slope must be avoided as
        # particles should drift downwind.
        # The problem arises because of high error variances (see e.g. PIW-1).
        downwind_slope = ones*self.leewayprop[object_type]['DWSLOPE']
        downwind_offset = ones*self.leewayprop[object_type]['DWOFFSET']
        dwstd = self.leewayprop[object_type]['DWSTD']
        rdw = np.zeros(number)
        epsdw = np.zeros(number)
        for i in range(number):
            rdw[i] = np.random.randn(1)
            epsdw[i] = rdw[i]*dwstd
            # Avoid negative downwind slopes
            while downwind_slope[i] + epsdw[i]/20.0 < 0.0:
                rdw[i] = np.random.randn(1)
                epsdw[i] = rdw[i]*dwstd
        downwind_eps = epsdw
        # NB
        # downwind_eps = np.zeros(number)

        # Crosswind leeway properties
        rcw = np.random.randn(number)
        crosswind_slope = np.zeros(number)
        crosswind_offset = np.zeros(number)
        crosswind_eps = np.zeros(number)
        crosswind_slope[orientation == RIGHT] = \
            self.leewayprop[object_type]['CWRSLOPE']
        crosswind_slope[orientation == LEFT] = \
            self.leewayprop[object_type]['CWLSLOPE']
        crosswind_offset[orientation == RIGHT] = \
            self.leewayprop[object_type]['CWROFFSET']
        crosswind_offset[orientation == LEFT] = \
            self.leewayprop[object_type]['CWLOFFSET']
        crosswind_eps[orientation == RIGHT] = \
            rcw[orientation == RIGHT] * \
            self.leewayprop[object_type]['CWRSTD']
        crosswind_eps[orientation == LEFT] = \
            rcw[orientation == LEFT] * \
            self.leewayprop[object_type]['CWLSTD']

        # NB
        # crosswind_eps = np.zeros(number)

        # Store seed data for ASCII format output
        if hasattr(self, 'seed_cone_arguments'):
            self.ascii = self.seed_cone_arguments
        else:
            self.ascii = {'lon': lon, 'lat': lat,
                          'radius': kwargs['radius'] if 'radius' in kwargs else 0,
                          'number': number, 'time': kwargs['time']}

        # Call general seed_elements function of OpenDriftSimulation class
        # with the specific values calculated
        super(Leeway, self).seed_elements(lon, lat,
            orientation=orientation, object_type=object_type,
            downwind_slope=downwind_slope,
            crosswind_slope=crosswind_slope,
            downwind_offset=downwind_offset,
            crosswind_offset=crosswind_offset,
            downwind_eps=downwind_eps, crosswind_eps=crosswind_eps,
            **kwargs)

    def list_object_categories(self, substr=None):
        '''Display leeway categories to screen

        Print only objects containing 'substr', if specified'''

        for i, p in enumerate(self.leewayprop):
            description = self.leewayprop[p]['Description']
            objkey = self.leewayprop[p]['OBJKEY']
            if substr is not None:
                if substr.lower() not in description.lower() + objkey.lower():
                    continue
            print('%i %s %s' % (i+1, objkey, description))

    def update(self):
        """Update positions and properties of leeway particles."""

        windspeed = np.sqrt(self.environment.x_wind**2 +
                            self.environment.y_wind**2)
        # CCC update wind direction
        winddir = np.arctan2(self.environment.x_wind, self.environment.y_wind)

        # Move particles with the leeway CCC TODO
        downwind_leeway = ((self.elements.downwind_slope +
                            self.elements.downwind_eps/20.0)*windspeed +
                           self.elements.downwind_offset +
                           self.elements.downwind_eps/2.0)*.01  # In m/s
        crosswind_leeway = ((self.elements.crosswind_slope +
                            self.elements.crosswind_eps/20.0)*windspeed +
                            self.elements.crosswind_offset +
                            self.elements.crosswind_eps/2.0)*.01  # In m/s
        sinth = np.sin(winddir)
        costh = np.cos(winddir)
        y_leeway = downwind_leeway*costh+crosswind_leeway*sinth
        x_leeway = -downwind_leeway*sinth+crosswind_leeway*costh
        self.update_positions(-x_leeway, y_leeway)

        # Move particles with ambient current
        self.update_positions(self.environment.x_sea_water_velocity,
                              self.environment.y_sea_water_velocity)

        # Jibe elements randomly according to given probability
        jibe_rate = -np.log(1-self.elements.jibe_probability)/3600  # Hourly to instantaneous
        jp_per_timestep = 1 - np.exp(-jibe_rate*np.abs(self.time_step.total_seconds()))
        jib = jp_per_timestep > np.random.random(self.num_elements_active())
        self.elements.crosswind_slope[jib] = - self.elements.crosswind_slope[jib]
        self.elements.orientation[jib] = 1 - self.elements.orientation[jib]
        logger.debug('Jibing %i out of %i elements.' %
                      (np.sum(jib), self.num_elements_active()))

    def export_ascii(self, filename):
        '''Export output to ASCII format of original version'''

        try:
            f = open(filename, 'w')
        except:
            raise ValueError('Could not open file for writing: '
                             + filename)

        start_time = self.start_time
        for inp in ['lon', 'lat', 'radius', 'time']:
            if len(np.atleast_1d(self.ascii[inp])) == 1:
                self.ascii[inp] = [self.ascii[inp], self.ascii[inp]]
        f.write('# Drift simulation initiated [UTC]:\n')
        f.write('simDate simTime\n')
        f.write(self.start_time.strftime('%Y-%m-%d\t%H:%M:%S\t#\n'))
        f.write(
            '# Model version:\n'
            'modelVersion\n'
            ' 3.00\n'  # OpenDrift version
            '# Object class id & name:\n'
            'objectClassId  objectClassName\n')
        try:
            objtype = self.elements.object_type[0]
        except:
            objtype = self.elements_deactivated.object_type[0]
        f.write(' %i\t%s\n' % (objtype,
                self.leewayprop[objtype]['OBJKEY']))
        f.write(
            '# Seeding start time, position & radius:\n'
            'startDate\tstartTime\tstartLon\tstartLat\tstartRad\n')
        f.write('%s\t%s\t%s\t%s\n' % (
            self.ascii['time'][0], self.ascii['lon'][0],
            self.ascii['lat'][0], self.ascii['radius'][0]/1000.))
        f.write(
            '# Seeding end time, position & radius:\n'
            'endDate\tendTime\tendLon\tendLat\tendRad\n')
        f.write('%s\t%s\t%s\t%s\n' % (
            self.ascii['time'][1], self.ascii['lon'][1],
            self.ascii['lat'][1], self.ascii['radius'][1]/1000.))
        seedDuration = (self.ascii['time'][1]-self.ascii['time'][0]).total_seconds()/60.
        seedSteps=seedDuration/(self.time_step_output.total_seconds()/60.)
        seedSteps = np.maximum(1, seedSteps)
        f.write(
            '# Duration of seeding [min] & [timesteps]:\n'
            'seedDuration   seedSteps\n'
            '    %i      %i\n'
            '# Length of timestep [min]:\n'
            'timeStep\n' % (seedDuration, seedSteps))
        f.write('%i\n' % (self.time_step_output.total_seconds()/60.))
        f.write(
            '# Length of model simulation [min] & [timesteps]:\n'
            'simLength  simSteps\n')
        f.write('%i\t%i\n' % ((self.time - self.start_time).total_seconds()/60., self.steps_output))

        f.write(
            '# Total no of seeded particles:\n'
            'seedTotal\n'
            ' %s\n' % (self.num_elements_activated()))
        f.write(
            '# Particles seeded per timestep:\n'
            'seedRate\n'
            ' %i\n' % (self.num_elements_activated()/seedSteps))

        index_of_first, index_of_last = \
            self.index_of_activation_and_deactivation()

        beforeseeded = 0
        lons, statuss = self.get_property('lon')
        lats, statuss = self.get_property('lat')
        orientations, statuss = self.get_property('orientation')
        for step in range(self.steps_output):
            lon = lons[step,:]
            lat = lats[step,:]
            orientation = orientations[step,:]
            status = statuss[step,:]
            if sum(status==0) == 0:  # All elements deactivated: using last position
                lon = lons[step-1,:]
                lat = lats[step-1,:]
                orientation = orientations[step-1,:]
                status = statuss[step-1,:]
            num_active = np.sum(~status.mask)
            status[status.mask] = 41  # seeded on land
            lon[status.mask] = 0
            lat[status.mask] = 0
            orientation[status.mask] = 0
            status[status==0] = 11 # active
            status[status==1] = 41 # stranded
            ID = np.arange(0, num_active+1)
            f.write('\n# Date [UTC]:\nnowDate   nowTime\n')
            f.write((self.start_time + self.time_step_output*step).strftime('%Y-%m-%d\t%H:%M:%S\n'))
            f.write('# Time passed [min] & [timesteps], now seeded, seeded so far:\ntimePassed  nStep   nowSeeded   nSeeded\n')
            f.write('  %i\t%i\t%i\t%i\n' % (
                (self.time_step_output*step).total_seconds()/60,
                 step+1, num_active-beforeseeded, num_active))
            beforeseeded = num_active
            f.write('# Mean position:\nmeanLon meanLat\n')
            f.write('%f\t%f\n' % (np.mean(lon[status==11]),
                                  np.mean(lat[status==11])))
            f.write('# Particle data:\n')
            f.write('id  lon lat state   age orientation\n')
            age_minutes = self.time_step_output.total_seconds()*(
                           step - index_of_first)/60
            age_minutes[age_minutes<0] = 0
            for i in range(num_active):
                f.write('%i\t%s\t%s\t%i\t%i\t%i\n' % (i+1,
                        lon[i], lat[i], status[i], age_minutes[i],
                        orientation[i]))

        f.close()

    def _substance_name(self):
        # TODO: find a better algorithm to return name of object category
        if self.history is not None:
            object_type = self.history['object_type'][0,0]
            if not np.isfinite(object_type):  # For backward simulations
                object_type = self.history['object_type'][-1,0]
        else:
            object_type = np.atleast_1d(self.elements_scheduled.object_type)[0]
        if np.isfinite(object_type):
            return self.leewayprop[object_type]['OBJKEY']
        else:
            return ''



# Som valid (but extreme) ranges for checking that values are reasonable
standard_names = {
    'x_wind': {'valid_min': -50, 'valid_max': 50, 'units': 'm/s',
               'long_name': 'Component of wind along x-direction '
                            '(eastwards if projection is lonlat/Mercator)'},
    'y_wind': {'valid_min': -50, 'valid_max': 50, 'units': 'm/s',
        'long_name': 'Component of wind along y-direction (northwards if '
                      'projection is lonlat/Mercator)'},

    'x_sea_water_velocity': {'valid_min': -15, 'valid_max': 15,
        'units': 'm/s',
        'long_name': 'Component of ocean current along x-direction '
                     ' (eastwards if projection is lonlat/Mercator)'},
    'y_sea_water_velocity': {'valid_min': -15, 'valid_max': 15, 'units': 'm/s',

        'long_name': 'Component of ocean current along y-direction '
                      '(northwards if projection is lonlat/Mercator)'},
    'land_binary_mask': {'valid_min': 0, 'valid_max': 1,
                         'long_name': '1 is land, 0 is sea'},
    'sea_floor_depth_below_sea_level': {'valid_min': 0, 'valid_max': 12000,
                         'long_name': 'Depth of seafloor'},
    'ocean_vertical_diffusivity': {'valid_min': 0, 'valid_max': 1}}

# Identify x-y vector components/pairs for rotation (NB: not east-west pairs!)
vector_pairs_xy = [
    ['x_wind', 'y_wind'],
    ['sea_ice_x_velocity', 'sea_ice_y_velocity'],
    ['x_sea_water_velocity', 'y_sea_water_velocity'],
    ['sea_surface_wave_stokes_drift_x_velocity',
     'sea_surface_wave_stokes_drift_y_velocity']
    ]



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway
# Copyright 2020, Gaute Hope, MET Norway

import sys
import logging
logger = logging.getLogger(__name__)
from abc import abstractmethod, ABCMeta
from datetime import datetime, timedelta

import numpy as np
import pyproj
import matplotlib.pyplot as plt
from matplotlib.patches import Polygon
import cartopy.crs as ccrs
import cartopy.feature as cfeature

from .structured import StructuredReader
from .unstructured import UnstructuredReader
from .continuous import ContinuousReader
from .variables import Variables
from .consts import *
from ..operators.ops import Combine, Filter

from opendrift.readers.interpolation import ReaderBlock

class BaseReader(Variables, Combine, Filter):
    """
    An abstract reader. Implementors provide a method to read data and specify how it is interpolated.

    This class inherits :class:`.variables.Variables` which inherits :class:`.variables.ReaderDomain`. `ReaderDomain` is responsible for the extent and domain of the reader, including checking for out-of-bounds and projection conversion. `Variables` is responsible for returning interpolated data at the requests positions or profiles. Apart from coercing the returned data into the right type for :py:mod:`opendrift.models.basemodel`, it defines the abstract interface to :meth:`.variables.Variables._get_variables_interpolated_` which reader-implementations must provide (_usually_ through one of the main reader-types, see: :py:mod:`opendrift.readers`).

    .. seealso::

        :py:mod:`opendrift.readers`.

        :py:mod:`.variables`.
    """

    __metaclass__ = ABCMeta

    verticalbuffer = 1  # To be overridden by application as needed

    # Mapping variable names, e.g. from east-north to x-y, temporarily
    # presuming coordinate system then is lon-lat for equivalence
    variable_aliases = {
        'sea_water_potential_temperature': 'sea_water_temperature',
        'x_wind_10m': 'x_wind',
        'y_wind_10m': 'y_wind',
        'surface_wind_speed': 'wind_speed',
        'surface_wind_direction': 'wind_from_direction',
        'wind_direction': 'wind_from_direction',
        'sea_water_x_velocity': 'x_sea_water_velocity',
        'sea_water_y_velocity': 'y_sea_water_velocity',
        'x_sea_ice_velocity': 'sea_ice_x_velocity',
        'y_sea_ice_velocity': 'sea_ice_y_velocity',
        'barotropic_sea_water_x_velocity': 'sea_ice_x_velocity',
        'barotropic_sea_water_y_velocity': 'sea_ice_y_velocity',
        'salinity_vertical_diffusion_coefficient' : 'ocean_vertical_diffusivity',
        'ocean_mixed_layer_thickness_defined_by_sigma_theta': 'ocean_mixed_layer_thickness',
        'sea_floor_depth_below_sea_surface' : 'sea_floor_depth_below_sea_level',
        'sea_floor_depth_below_geoid' : 'sea_floor_depth_below_sea_level',
        'mass_concentration_of_suspended_matter_in_sea_water' : 'spm'
        }

    xy2eastnorth_mapping = {
        'x_sea_water_velocity': ['eastward_sea_water_velocity', 'surface_eastward_sea_water_velocity',
                                 'eastward_current_velocity', 'eastward_tidal_current',
                                 'eastward_ekman_current_velocity', 'eastward_geostrophic_current_velocity',
                                 'eastward_eulerian_current_velocity', 'surface_geostrophic_eastward_sea_water_velocity',
                                 'surface_geostrophic_eastward_sea_water_velocity_assuming_sea_level_for_geoid',
                                 'surface_eastward_geostrophic_sea_water_velocity_assuming_sea_level_for_geoid'],
        'y_sea_water_velocity': ['northward_sea_water_velocity', 'surface_northward_sea_water_velocity',
                                 'northward_current_velocity', 'northward_tidal_current',
                                 'northward_ekman_current_velocity', 'northward_geostrophic_current_velocity',
                                 'northward_eulerian_current_velocity', 'surface_geostrophic_northward_sea_water_velocity',
                                 'surface_geostrophic_northward_sea_water_velocity_assuming_sea_level_for_geoid',
                                 'surface_northward_geostrophic_sea_water_velocity_assuming_sea_level_for_geoid'],
        'x_wind': 'eastward_wind', 'y_wind': 'northward_wind'}

    def __init__(self):
        """Common constructor for all readers"""
        super().__init__()

        self.always_valid = False  # Set to True if a single field should
                                   # be valid at all times

        self.is_lazy = False  # Generally False

        # Set projection for coordinate transformations
        if self.proj is None:
            if self.proj4 is not None:
                try:
                    self.proj = pyproj.Proj(self.proj4)
                except Exception as ex:
                    logger.exception(ex)
                    # Workaround for proj-issue with zero flattening:
                    # https://github.com/OSGeo/proj.4/issues/1191
                    origproj4 = self.proj4
                    self.proj4 = self.proj4.replace('+e=0.0', '')
                    self.proj4 = self.proj4.replace('+e=0', '')
                    self.proj4 = self.proj4.replace('+f=0.0', '')
                    self.proj4 = self.proj4.replace('+f=0', '')
                    if origproj4 != self.proj4:
                        logger.warning('Removing flattening parameter from proj4; %s -> %s' % (origproj4, self.proj4))
                    self.proj = pyproj.Proj(self.proj4)
            else:
                raise Exception("No projection or CRS information available")

        # Check if there are holes in time domain
        if self.start_time is not None and self.time_step is not None:# and len(self.times) > 1:
            self.expected_time_steps = (
                self.end_time - self.start_time).total_seconds() / (
                self.time_step.total_seconds()) + 1
            if self.times is not None:
                self.missing_time_steps = self.expected_time_steps - \
                    len(self.times)
            else:
                self.missing_time_steps = 0
            self.actual_time_steps = self.expected_time_steps - \
                self.missing_time_steps

        # Making sure start_time is datetime, and not cftime object
        if self.start_time is not None:
             self.start_time = datetime(self.start_time.year, self.start_time.month,
                                   self.start_time.day, self.start_time.hour,
                                   self.start_time.minute, self.start_time.second)

        # Calculate shape (size) of domain
        try:
            numx = (self.xmax - self.xmin)/self.delta_x + 1
            numy = (self.ymax - self.ymin)/self.delta_y + 1
            self.shape = (int(numx), int(numy))
        except:
            self.shape = None

        self.set_buffer_size(max_speed = 5.)

        # Check if there are east/north-oriented vectors
        for var in self.variables:
            for xvar, eastnorthvar in self.xy2eastnorth_mapping.items():
                if xvar in self.variables:
                    continue  # We have both x/y and east/north components
                if var in eastnorthvar:
                    logger.info('Variable %s will be rotated from %s' % (xvar, var))
                    self.variables.append(xvar)
                    if not hasattr(self, 'rotate_mapping'):
                        self.rotate_mapping = {}
                    self.rotate_mapping[xvar] = var

        # Adding variables which may be derived from existing ones
        for m in self.environment_mappings:
            em = self.environment_mappings[m]
            if em['active'] is False:
                logger.debug('Variable mapping: %s -> %s is not activated' % (em['input'], em['output']))
                continue
            self.activate_environment_mapping(m)

    def y_is_north(self):
        if self.proj.crs.is_geographic or '+proj=merc' in self.proj.srs:
            return True
        else:
            return False

    def prepare(self, extent, start_time, end_time, max_speed):
        """Prepare reader for given simulation coverage in time and space."""
        logger.debug('Nothing more to prepare for ' + self.name)
        pass  # to be overriden by specific readers

    def rotate_variable_dict(self, variables, proj_from='+proj=latlong', proj_to=None):
        for vectorpair in vector_pairs_xy:
            if vectorpair[0] in self.rotate_mapping and vectorpair[0] in variables.keys():
                if proj_to is None:
                    proj_to = self.proj
                logger.debug('Rotating vector from east/north to xy orientation: ' + str(vectorpair))
                variables[vectorpair[0]], variables[vectorpair[1]] = self.rotate_vectors(
                    variables['x'], variables['y'],
                    variables[vectorpair[0]], variables[vectorpair[1]],
                    proj_from, proj_to)

    def index_of_closest_z(self, requested_z):
        """Return (internal) index of z closest to requested z.

        Thickness of layers (of ocean model) are not assumed to be constant.
        """
        ind_z = [np.abs(np.subtract.outer(
            self.z, requested_z)).argmin(0)]
        return ind_z, self.z[ind_z]

    def indices_min_max_z(self, z):
        """
        Return min and max indices of internal vertical dimension,
        covering the requested vertical positions.
        Needed when block is requested (True).

        Arguments:
            z: ndarray of floats, in meters
        """
        minIndex = (self.z <= z.min()).argmin() - 1
        maxIndex = (self.z >= z.max()).argmax()
        return minIndex, maxIndex

    def performance(self):
        '''Report the time spent on various tasks'''
        outStr = ''
        if hasattr(self, 'timing'):
            for cat, time in self.timing.items():
                time = str(time)[0:str(time).find('.') + 2]
                outStr += '%10s  %s\n' % (time, cat)
        return outStr

    def clip_boundary_pixels(self, numpix):
        '''Trim some (potentially bad) pixels along boundary'''
        logger.info('Trimming %i pixels from boundary' % numpix)
        self.xmin = self.xmin+numpix*self.delta_x
        self.xmax = self.xmax-numpix*self.delta_x
        self.ymin = self.ymin+numpix*self.delta_y
        self.ymax = self.ymax-numpix*self.delta_y
        self.shape = tuple([self.shape[0]-2*numpix,
                            self.shape[1]-2*numpix])
        self.clipped = numpix

    def plot(self, variable=None, vmin=None, vmax=None,
             filename=None, title=None, buffer=1, lscale='auto'):
        """Plot geographical coverage of reader."""

        fig = plt.figure()

        if self.global_coverage():
            lonmin=-180
            lonmax=180
            latmin=-89
            latmax=89
        else:
            corners = self.xy2lonlat([self.xmin, self.xmin, self.xmax, self.xmax],
                                     [self.ymax, self.ymin, self.ymax, self.ymin])
            lonmin = np.min(corners[0]) - buffer*2
            lonmax = np.max(corners[0]) + buffer*2
            latmin = np.min(corners[1]) - buffer
            latmax = np.max(corners[1]) + buffer

        # Initialise map
        if not self.global_coverage():
            # Stereographic projection centred on domain, if small domain
            x0 = (self.xmin + self.xmax) / 2
            y0 = (self.ymin + self.ymax) / 2
            lon0, lat0 = self.xy2lonlat(x0, y0)
            lon0 = lon0[0]
            lat0 = lat0[0]
            sp = ccrs.Stereographic(central_longitude=lon0, central_latitude=lat0)
            latmax = np.maximum(latmax, lat0)
            latmin = np.minimum(latmin, lat0)
        else:
            # Global map if reader domain is large
            sp = ccrs.Mercator()

        ax = fig.add_subplot(1, 1, 1, projection=sp)

        if lscale == 'auto':  # Custom lscale - this should be generalized to Basemodel also
            s = cfeature.AdaptiveScaler('coarse',
                (('low', 100), ('intermediate', 20), ('high', 5), ('full', 1)))
            lscale = s.scale_from_extent([lonmin, lonmax, latmin, latmax])

        # GSHHS coastlines
        f = cfeature.GSHHSFeature(scale=lscale, levels=[1])
        f._geometries_cache = {}
        ax.add_geometries(
            #f.intersecting_geometries([lonmin, lonmax, latmin, latmax]),
            f.geometries(),
            ccrs.PlateCarree(),
            facecolor=cfeature.COLORS['land'],
            edgecolor='black')

        gl = ax.gridlines(ccrs.PlateCarree())
        gl.top_labels = False

        # Get boundary
        if self.global_coverage():
            lon = np.array([-180, 180, 180, -180, -180])
            lat = np.array([-89, -89, 89, 89, -89])
        else:
            npoints = 10  # points per side
            x = np.array([])
            y = np.array([])
            x = np.concatenate((x, np.linspace(self.xmin, self.xmax, npoints)))
            y = np.concatenate((y, [self.ymin]*npoints))
            x = np.concatenate((x, [self.xmax]*npoints))
            y = np.concatenate((y, np.linspace(self.ymin, self.ymax, npoints)))
            x = np.concatenate((x, np.linspace(self.xmax, self.xmin, npoints)))
            y = np.concatenate((y, [self.ymax]*npoints))
            x = np.concatenate((x, [self.xmin]*npoints))
            y = np.concatenate((y, np.linspace(self.ymax, self.ymin, npoints)))
            # from x/y vectors create a Patch to be added to map
            lon, lat = self.xy2lonlat(x, y)
            lat[lat>89] = 89.
            lat[lat<-89] = -89.
        p = sp.transform_points(ccrs.PlateCarree(), lon, lat)
        xsp = p[:, 0]
        ysp = p[:, 1]

        if variable is None:
            boundary = Polygon(list(zip(xsp, ysp)), alpha=0.5, ec='k', fc='b',
                               zorder=100)
            ax.add_patch(boundary)
            if self.global_coverage():
                ax.set_global()
            else:
                buf = (xsp.max()-xsp.min())*.1  # Some whitespace around polygon
                ax.set_extent([xsp.min()-buf, xsp.max()+buf, ysp.min()-buf, ysp.max()+buf], crs=sp)
        if title is None:
            plt.title(self.name)
        else:
            plt.title(title)
        plt.xlabel('Time coverage: %s to %s' %
                   (self.start_time, self.end_time))

        if variable is not None:
            rx = np.array([self.xmin, self.xmax])
            ry = np.array([self.ymin, self.ymax])
            if variable in self.derived_variables:
                data = self.get_variables(self.derived_variables[variable], self.start_time, rx, ry)
                self.__calculate_derived_environment_variables__(data)
            else:
                data = self.get_variables(variable, self.start_time, rx, ry)
            rx, ry = np.meshgrid(data['x'], data['y'])
            rx = np.float32(rx)
            ry = np.float32(ry)
            rlon, rlat = self.xy2lonlat(rx, ry)
            data[variable] = np.ma.masked_invalid(data[variable])

            if data[variable].ndim > 2:
                logger.warning('Ensemble data, plotting only first member')
                data[variable] = data[variable][0,:,:]
            if self.global_coverage():
                mappable = ax.pcolormesh(rlon, rlat, data[variable], vmin=vmin, vmax=vmax,
                                         transform=ccrs.PlateCarree())
            else:
                p = sp.transform_points(ccrs.PlateCarree(), rlon, rlat)
                mapx = p[:,:,0]
                mapy = p[:,:,1]
                mappable = ax.pcolormesh(mapx, mapy, data[variable], vmin=vmin, vmax=vmax)

            cbar = fig.colorbar(mappable, orientation='horizontal', pad=.05, aspect=30, shrink=.4)
            cbar.set_label(variable)

        try:  # Activate figure zooming
            mng = plt.get_current_fig_manager()
            mng.toolbar.zoom()
        except:
            pass

        if filename is not None:
            plt.savefig(filename)
            plt.close()
        else:
            plt.show()

    def get_timeseries_at_position(self, lon, lat, variables=None,
                                   start_time=None, end_time=None, times=None):
        """ Get timeseries of variables from this reader at given position.
        """

        if times is None:
            if start_time is None:
                start_time = self.start_time
            if end_time is None:
                end_time = self.end_time
            times = [t for t in self.times if t >= start_time and t<= end_time]

        if variables is None:
            variables = self.variables

        # if len(self.covers_positions(lon=lon, lat=lat)[0]) == 0:
        #     return None

        lon = np.atleast_1d(lon)
        lat = np.atleast_1d(lat)
        if len(lon) == 1:
            lon = lon[0]*np.ones(len(times))
            lat = lat[0]*np.ones(len(times))

        data = {'time': times}
        for var in variables:
            data[var] = np.zeros(len(times))

        for i, time in enumerate(times):
            closest_time = min(self.times, key=lambda d: abs(d - time))
            d = self.get_variables_interpolated(
                lon=np.atleast_1d(lon[i]), lat=np.atleast_1d(lat[i]), z=np.atleast_1d(0),
                time=closest_time, variables=variables, rotate_to_proj='+proj=latlong')[0]
            for var in variables:
                data[var][i] = d[var][0]

        return data

    def shift_start_time(self, start_time):
        """Shift the time coverage of reader to match given start_time"""
        if self.start_time is None:
            logger.warning('Reader has no time coverage, cannot shift times')
            return
        shift_time = self.start_time - start_time
        self.start_time = start_time
        self.end_time = self.end_time - shift_time
        self.times = [t - shift_time for t in self.times]



"""
warnings.py

All the warnings
"""

# FIXME: it would be better to have some structure to he warning codes

WARNINGS = {
    "W001":
        "Record name: {} is not very descriptive",
    "W002":
        "Record has no product type",
    "W003":
        '"{}" is not a valid product type. Options are: {}',
    "W004":
        "No api value provided",
    "W005":
        "API value: {api} seems unlikely",
    "W011":
        "{} date format: {} is invalid: {}",
    "W006":
        "No density values provided",
    "W007":
        "No distillation data provided",
    "W008":
        "No reference year provided",
    "W009":
        "Distillation fraction recovered is missing or invalid",
    "W010":
        "Temperature: {} is close to {} -- looks like it could be a K to C conversion error",
}

WARNINGS = {code: (code + ": " + msg) for code, msg in WARNINGS.items()}



from abc import abstractmethod
import numpy as np
from .variables import Variables
from .consts import vector_pairs_xy
import logging
logger = logging.getLogger(__name__)

class ContinuousReader(Variables):
    """
    A continuous (in space and time) reader, able to provide
    exact values at any desired point (within bounds). This reader type is
    suitable for constant readers, analytical readers, or readers that are
    static and continuous within the valid domain (e.g. the landmask reader).

    .. seealso::

        :py:mod:`opendrift.readers`
    """

    @abstractmethod
    def get_variables(self, variables, time=None, x=None, y=None, z=None):
        """
        Obtain and return values of the requested variables at all positions
        (x, y, z) closest to given time.

        Returns:

            Dictionary with arrays of length len(x) with values at exact positions x, y and z.
        """

    def _get_variables_interpolated_(self, variables, profiles,
                                   profiles_depth, time,
                                   reader_x, reader_y, z):

        env = self.get_variables(variables, time, reader_x, reader_y, z)

        logger.debug('Fetched env-before')
        env_profiles = None
        if profiles is not None:
            # Copying data from environment to vertical profiles
            env_profiles = {'z': profiles_depth}
            for var in profiles:
                env_profiles[var] = np.ma.array([env[var], env[var]])

        return env, env_profiles




"""
cleanup modules

all cleanup classes are registered here

so far:
 - compute density from API

possible things to cleanup:

* normalizing units for various things

* putting distillation cuts in order

"""

# you need to import a cleanup class here to get it registered

from .cleanup import Cleanup

from .density import FixAPI

ALL_CLEANUPS = []
CLEANUP_MAPPING = {}  # mapping to get cleanup object from its ID

for name, obj in dict(locals()).items():
    # print("in init, obj:")
    # print(obj)
    if (not name.startswith("_")
        and isinstance(obj, type)
        and issubclass(obj, Cleanup)
        and (obj is not Cleanup)):

        ALL_CLEANUPS.append(obj)
        CLEANUP_MAPPING[obj.ID] = obj









"""
Readers
=======

Readers are responsible for providing Opendrift with data about the
`enviornment` of the drifting particles or elements.

All readers descend from :class:`.basereader.BaseReader`. A reader generally also descends from one of the few general classes of readers. When writing a new reader consider which one fits your input data best:

    * :class:`.basereader.continuous.ContinuousReader`
    * :class:`.basereader.structured.StructuredReader`
    * :class:`.basereader.unstructured.UnstructuredReader`

The `ContinuousReader` is suited for data that can be defined at any point within the domain, or if the reader does its own interpolation internally. E.g. a :class:`synthetic eddy <.reader_ArtificialOceanEddy.Reader>`, or a :class:`constant <.reader_constant.Reader>`. The `StructuredReader` aids in interpolation when creating a reader of data on a :class:`regular grid <.reader_netCDF_CF_generic.Reader>`, while the `UnstructuredReader` provides the basics for :class:`irregularily gridded data <.reader_netCDF_CF_unstructured.Reader>` (e.g. finite volume models).

.. seealso::

    See the reader-types or reader-implementations for more details.

    See :class:`.basereader.BaseReader` for how readers work internally.
"""

import importlib
import logging; logger = logging.getLogger(__name__)
import glob
import json
import opendrift

def reader_from_url(url, timeout=10):
    '''Make readers from URLs or paths to datasets'''

    if isinstance(url, list):
        return [reader_from_url(u) for u in url]

    try:  # Initialise reader from JSON string
        j = json.loads(url)
        try:
            reader_module = importlib.import_module(
                    'opendrift.readers.' + j['reader'])
            reader = getattr(reader_module, 'Reader')
            del j['reader']
            reader = reader(**j)
            return reader
        except Exception as e:
            logger.warning('Creating reader from JSON failed:')
            logger.warning(e)
    except:
        pass

    if len(glob.glob(url)) == 0:  # Check if this is a URL, and not giving timeout
        import requests
        try:
            resp = requests.get(url, timeout=timeout)
        except requests.exceptions.MissingSchema:
            logger.info('Neither a file or a URL: ' + url)
            return None
        except:
            logger.info('Connection error for ' + url)
            return None

    reader_modules = ['reader_netCDF_CF_generic',
                      'reader_ROMS_native',
                      'reader_grib']

    for rm in reader_modules:
        reader_module = importlib.import_module('opendrift.readers.' + rm)
        try:
            r = reader_module.Reader(url)
            return r
        except Exception as e:
            print('Could not open %s with %s' % (url, rm))

    return None  # No readers worked



'''
    Classes for storing measured values within an Oil record
'''
from dataclasses import dataclass, field

from ..common.utilities import dataclass_to_json, JSON_List
from ..common.measurement import (Time,
                                  Temperature,
                                  MassFraction,
                                  VolumeFraction,
                                  MassOrVolumeFraction,
                                  Concentration,
                                  Unitless,
                                  Dimensionless,
                                  Density,
                                  DynamicViscosity,
                                  KinematicViscosity,
                                  InterfacialTension,
                                  Pressure,
                                  AngularVelocity)

from ..common.validators import EnumValidator
from .validation.warnings import WARNINGS
from .validation.errors import ERRORS


@dataclass_to_json
@dataclass
class DistCut:
    fraction: MassOrVolumeFraction
    vapor_temp: Temperature


class DistCutList(JSON_List):
    item_type = DistCut


@dataclass_to_json
@dataclass
class Distillation:
    type: str = None
    method: str = None
    end_point: Temperature = None
    fraction_recovered: MassOrVolumeFraction = None
    cuts: DistCutList = field(default_factory=DistCutList)

    def validate(self):
        msgs = []
        if self.cuts:  # only need to validate if there are cuts
            msgs.extend(EnumValidator({"mass fraction", "volume fraction"},
                                      ERRORS["E032"],
                                      case_insensitive=True)(self.type))
            if self.fraction_recovered is None:
                msgs.append(WARNINGS["W009"])
            else:
                frac_recov = self.fraction_recovered.converted_to("fraction")
                if frac_recov.value is not None:
                    val = frac_recov.value
                elif frac_recov.max_value is not None:
                    val = frac_recov.max_value

                else:
                    val = None
                    msgs.append(WARNINGS["W009"])
                if val is not None:
                    if not (0.0 <= val <= 1.0):
                        msgs.append(ERRORS["E041"]
                                    .format("distillation fraction recovered",
                                            val))

            for cut in self.cuts:
                frac = cut.fraction.converted_to('fraction').value
                if not (0.0 <= frac <= 1.0):
                    msgs.append(ERRORS["E041"]
                                .format("distillation fraction", frac))
                vt = cut.vapor_temp.convert_to('C').value
                if vt < -100.0:
                    t = f"{cut.vapor_temp.value:.2f} {cut.vapor_temp.unit}"
                    msgs.append(ERRORS["E040"]
                                .format("distillation vapor temp", t))

        return msgs



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2021, Julien Moreau, Plastic@Bay CIC

from datetime import timedelta, datetime
import numpy as np
import logging; logger = logging.getLogger(__name__)
from opendrift.models.oceandrift import Lagrangian3DArray, OceanDrift
from opendrift.models.physics_methods import hour_angle, solar_elevation

class SeaLiceElement(Lagrangian3DArray):
    """
    Extending Lagrangian3DArray with specific properties for larval and
    juvenile stages of sea lice into super individuals
    """

    variables = Lagrangian3DArray.add_variables([
        ('LicePerFish',{'dtype':np.float32,
                        'units':'',
                        'default':0.5}),
        ('AvFishW8',{'dtype':np.float32,
                    'units':'kg',
                    'default':4.5}),
        ('particle_biomass',{'dtype': np.float32,
                            'units': 'kg',
                            'default': 1000.}),
        ('hatched',{'dtype': np.float32,
                            'units': '',
                            'default': 0.}),
        ('nauplii', {'dtype': np.float32,
                            'units': '',
                            'default': 0.}),
        ('copepodid', {'dtype': np.float32,
                     'units': '',
                     'default': 0.}),
        ('dead',  {'dtype': np.float32,
                     'units': '',
                     'default': 0.}),
        ('eliminated',{'dtype': np.int8,
                     'units': '',
                     'default': 0}),
        ('degree_days', {'dtype': np.float32,
                     'units': '',
                    'default': 0}), #range 40-170
        ('safe_salinity_above', {'dtype': np.int8,
                     'units': '',
                    'default': 0}),
        ('temperature_above', {'dtype': np.float32,
                     'units': '',
                    'default': 10}),
        ('temperature_below', {'dtype': np.float32,
                     'units': '',
                    'default': 10}),
        ('light', {'dtype': np.float32,
                     'units': 'µmol photon s−1 m−2',
                     'default': 0.}),
                      ])


class SeaLice(OceanDrift):
    """
    Particle trajectory model based on the OpenDrift framework.
    Developed by Julien Moreau (Plastics@Bay CIC)
    Generic module for particles that are subject to vertical turbulent
    mixing with the possibility for positive or negative buoyancy
    Particles are sea-lice (Lepeophtheirus salmonis).
    """

    ElementType = SeaLiceElement

    required_variables = {
        'x_sea_water_velocity': {'fallback': 0},
        'y_sea_water_velocity': {'fallback': 0},
        # 'sea_surface_wave_significant_height': {'fallback': 0},
        # 'x_wind': {'fallback': 0},
        # 'y_wind': {'fallback': 0},
        'land_binary_mask': {'fallback': None},
        'sea_floor_depth_below_sea_level': {'fallback': 50},
        'surface_net_downward_radiative_flux':{'fallback': 0},
        'ocean_vertical_diffusivity': {'fallback': 0.01},
        'sea_water_temperature': {'fallback': 10},
        'sea_water_salinity': {'fallback': 34}
    }

    # required_profiles_z_range = [0, -50]  # The depth range (in m) which profiles should cover

    def __init__(self,*args, **kwargs):

        # Calling general constructor of parent class
        # super(SeaLice, self).__init__(*args, **kwargs)
        super().__init__(*args, **kwargs)

        # Configuration options
        self._add_config({
            'general:duration':{'type':'float', 'default':0.,
                                'min': None, 'max': None, 'units': 'seconds',
                                'description': 'Experiment time in seconds',
                                'level': self.CONFIG_LEVEL_ESSENTIAL},
            'lice:seeding_time_step':{'type':'float', 'default':None,
                                'min': None, 'max': None, 'units': 'seconds',
                                'description': 'Time between particle release',
                                'level': self.CONFIG_LEVEL_ESSENTIAL},
            'lice:death_rate':{'type':'float', 'default':0.01/3600,
                                'min': 0., 'max': None, 'units': 's-1',
                                'description': 'Rate of Larvae death per seconds',
                                'level': self.CONFIG_LEVEL_BASIC},
            'lice:maturation_rate':{'type':'float', 'default':0.1/3600,
                                'min': 0., 'max': None, 'units': 's-1',
                                'description': 'Rate of Nauplii maturation in Copepodids',
                                'level': self.CONFIG_LEVEL_BASIC},
            'lice:maturity_date':{'type':'float', 'default':3.63,
                                'min': 0., 'max': None, 'units': 'days',
                                'description': 'Days to start maturing into Copepodids',
                                'level': self.CONFIG_LEVEL_BASIC},
            'lice:sinking_velocity':{'type':'float', 'default':0.00025,
                                'min': 0., 'max': 0.01, 'units': 'm.s-1',
                                'description': 'Larvae sinking velocity',
                                'level': self.CONFIG_LEVEL_BASIC},
            'lice:vertical_migration_speed':{'type':'float', 'default':0.00075,
                                'min': 0., 'max': 0.01, 'units': 'm.s-1',
                                'description': 'Larvae vertical speed',
                                'level': self.CONFIG_LEVEL_BASIC},
            'lice:freezing_salinity':{'type':'float', 'default':27.,
                                'min': 0., 'max': 35., 'units': 'PSU',
                                'description': 'Salinity immobilising larvae',
                                'level': self.CONFIG_LEVEL_BASIC},
            'lice:avoided_salinity':{'type':'float', 'default':32.,
                                'min': 0., 'max': 50., 'units': 'PSU',
                                'description': 'Salinity actively avoided',
                                'level': self.CONFIG_LEVEL_BASIC},
            'lice:nu':{'type':'float', 'default':500.,
                                'min': 100, 'max': 1000, 'units': 'nm',
                                'description': 'Wavelength used to calculate irradiance',
                                'level': self.CONFIG_LEVEL_ADVANCED},
            'lice:k_water':{'type':'float', 'default':0.2,
                                'min': -10., 'max': 0., 'units': '',
                                'description': 'coefficient of exponential decay of light in water',
                                'level': self.CONFIG_LEVEL_ADVANCED},
            'lice:Nauplii_light_trigger':{'type':'float', 'default':2.E-5,
                                'min': 0., 'max': 1., 'units': 'µmol photon  s−1 m−2',
                                'description': 'light detection threshold of Nauplii',
                                'level': self.CONFIG_LEVEL_BASIC},
            'lice:Copepodid_light_trigger':{'type':'float', 'default':0.392,
                                'min': 0.0, 'max': 1.0, 'units': 'µmol photon  s−1 m−2',
                                'description': 'light detection threshold of copepodids',
                                'level': self.CONFIG_LEVEL_BASIC},
            'lice:twilight':{'type':'float', 'default':15,
                                'min': 0., 'max': 90., 'units': 'degrees',
                                'description': 'angle below the horizon for twilight',
                                'level': self.CONFIG_LEVEL_ADVANCED},
            })

    def prepare_run(self):
        logger.info("preparing the run...")
        ### need to find a better way of initialising the variables...
        self.prefix="lice:"
        self.vertical_migration_speed=self.get_config(self.prefix+'vertical_migration_speed') \
                                        *self.time_step.total_seconds()
        self.sensing_distance=2*self.vertical_migration_speed
        self.tau= np.pi+2*np.deg2rad(self.get_config(self.prefix+'twilight')) #width of solar irridation distribution
        self.nu=self.get_config(self.prefix+'nu')
        self.sinking_velocity=self.get_config(self.prefix+'sinking_velocity') \
                                *self.time_step.total_seconds()
        self.freezing_salinity=self.get_config(self.prefix+'freezing_salinity')
        self.avoided_salinity=self.get_config(self.prefix+'avoided_salinity')
        self.k_water=self.get_config(self.prefix+'k_water')
        self.Nauplii_light_trigger=self.get_config(self.prefix+'Nauplii_light_trigger')
        self.Copepodid_light_trigger=self.get_config(self.prefix+'Copepodid_light_trigger')
        self.new_born()
        self.population()
        self.ref_date = datetime(1,1,1,0,0)

    def new_born(self):
        """
        Approach by Rittenhouse et al., (2016) [Rittenhouse, M.A., C.W. Revie
        and A. Hurford, 2016. A model for sea lice (Lepeophtheirus salmonis)
        dynamics in a seasonally changing environment. Energetics, 16, 8–16.,
        whereby the numbers of lice nauplii released every day (NP) are
        estimated according to the following equation:
        NP(t)=ηενA(t)
        where η = 592 is the number of eggs per clutch (Heuch et al., 2000;
        Rittenhouse et al., 2016),
        ε = 0.0476 is the egg string production rate per day (Heuch et al.,
        2000; Rittenhouse et al., 2016),
        ν = 0.6 is the hatching success i.e. the proportion of eggs which
        produce viable nauplii (Johnson and Albright, 1991),
        and A(t) is the total number of adult female lice on each farm,
        derived here from lice target levels and number of fish on site.
        """
        logger.debug("Calculating standard spawn of nauplii")
        Ne, nu = 592 , 0.6
        Eps=0.0476/timedelta(days=1).total_seconds()*self.get_config( \
                                                self.prefix+'seeding_time_step')
        self.spawn=Ne* Eps* nu

    def population(self):
        """
        Simulate the population evolution according to the biological parameters
        Parameters
        ----------
        t: numpy array (float)
        Time abscissa in seconds through the experiment
        self.Mat: int
        Age in timestep when the nauplii start to become adult copepodids
        maturation_rate: float
        Probability of maturation by time-step (t>= Mat)
        death_rate: float
        Probability of death by time-step

        Number of lice per particle when generated is standardised to 1
        Returns
        -------
        juv, adults, dead: np.arrays(int)
        Juveniles(Nauplii), adults(Copepodid), dead lices arrays through
        the duration of the experiment
        """
        logger.debug("Building global population model")
        death_rate=self.get_config(self.prefix+'death_rate')* self.time_step.total_seconds()
        maturation_rate=self.get_config(self.prefix+'maturation_rate')* self.time_step.total_seconds()
        duration = self.get_config('general:duration')/ self.time_step.total_seconds()
        Mat = int(np.ceil(self.get_config(self.prefix+'maturity_date')*24*3600/ \
                    self.time_step.total_seconds())) # maturity age in timestep
        t=np.arange(0,duration+1,dtype=np.int)
        self.juv=np.exp(-1*death_rate*t)
        self.juv[0]=1
        self.dead=1-self.juv
        self.adult=np.zeros_like(t, dtype=np.float)
        if Mat<duration:
            decayjuv=self.juv[Mat]*np.exp(-1*(maturation_rate+death_rate)*(t[t>=Mat]-Mat))
            self.juv[t>=Mat] =decayjuv
            self.adult[t>=Mat]=1-self.dead[t>=Mat]-self.juv[t>=Mat]

    def SI_pop(self):
        """
        distribute the age fractions in the particles
        """
        logger.debug("Aging the super_individuals")
        scaling = self.elements.particle_biomass*self.elements.LicePerFish/ \
                    self.elements.AvFishW8
        #identify new spawns
        New_release = self.elements.age_seconds<=self.time_step.total_seconds()
        self.elements.hatched[New_release]=self.spawn*scaling[New_release]

        self.elements.nauplii[New_release]=self.elements.hatched[New_release]
        Free = ~New_release
        time_in_step=(self.elements.age_seconds[Free]/ \
                    self.time_step.total_seconds()).astype(np.int)
        self.elements.nauplii[Free]=self.elements.hatched[Free]* \
                                    self.juv[time_in_step]
        self.elements.copepodid[Free]=self.elements.hatched[Free]* \
                                    self.adult[time_in_step]
        self.elements.dead[Free]=self.elements.hatched[Free]* \
                                    self.dead[time_in_step]
        # desactivate dead particles
        ## this test is not good for seeds with few lice
        Dying=self.elements.nauplii+ self.elements.copepodid<1
        self.elements.eliminated[Dying]=1
        self.deactivate_elements(self.elements.eliminated.astype(np.bool), reason="All dead")


    def sensing(self):
        """
        Lice sensing if above or bellow the conditions are better for them
        within a specified distance.
        moving the particles up and then down then back in their initial position
        """
        logger.debug("sensing temperature and salinity")
        Backup= np.copy(self.elements.z[:])
        self.elements.z +=self.sensing_distance
        self.elements.safe_salinity_above[self.environment.sea_water_salinity> \
                                            self.avoided_salinity]=1
        self.elements.temperature_above= self.environment.sea_water_temperature
        self.elements.z -= 2*self.sensing_distance
        self.elements.temperature_below= self.environment.sea_water_temperature
        self.elements.z = Backup

    def degree_days(self):
        """
        Calculate the degree days of a particles
        >>> under development <<<
        """
        logger.debug("accumulate degree_days **Experimental**")
        ### define active elements
        self.elements.degree_days+=self.environment.sea_water_temperature* \
                self.time_step.total_seconds()/timedelta(days=1).total_seconds()

    # def solar_noon(self):
    #     """
    #     Search solar noon at the longitude
    #     """
    #     self.ref_date=self.time.date.replace(hour=0,minute=0,second=0)
    #     Av_lon= np.mean(self.elements.lon)
    #     Av_lat= np.mean(self.elements.lat)
    #     minutes=np.arange(0,60*24, dtype=np.float32)
    #     angles = np.empty_like(minutes)
    #     for i in range(len(minutes)):
    #         angles[i]=hour_angle(self.ref_date+timedelta(minutes=float(minutes[i])),Av_lon)
    #
    #     solar_noon=o.ref_date+timedelta(minutes=float(minutes[np.argmin(np.abs(angles))]))
    #     self.noon_elevation=solar_elevation(solar_noon, Av_lon, Av_lat)


    def irradiance(self):
        """
        Distribute the daily energy from irradiance with a gaussian distribution.
        We use the twilight times for high sensitivity organisms
        Convert irradiance from W.m-2 to micromol photon.s-1.m-2
        https://www.berthold.com/en/bioanalytic/knowledge/faq/irradiance-to-photon-flux
        calculate the photon flux in the water according to exponential decay from the sea surface
        """
        self.elements.solar_angle=np.deg2rad(hour_angle(self.time,np.mean(self.elements.lon)))
        solar_coeff= np.sqrt(self.tau/(2*np.pi)*np.exp(-self.tau/2*self.elements.solar_angle**2))
        self.elements.light = solar_coeff* \
                        self.environment.surface_net_downward_radiative_flux* \
                        self.nu*0.00836* \
                        np.exp(self.k_water*self.elements.z)

    def depth_test(self):
        """
        The natural range of the larvae is 0-50m
        """
        self.elements.z[self.elements.z>0]=0
        self.elements.z[self.elements.z<-50]=-50

    def Lice_vertical_migration(self):
        """
        Make larval sea lice to migrate vertically according to salinity, light
        and temperature triggers.
        """
        ### all lice sink at the same speed
        self.elements.z -= self.sinking_velocity

        ### filter actively avoiding salt
        # Frozen=self.environment.sea_water_salinity<self.freezing_salinity
        Avoiding = np.logical_and((self.freezing_salinity< \
                    self.environment.sea_water_salinity),
                    (self.environment.sea_water_salinity<self.avoided_salinity))

        Normal_salt = self.environment.sea_water_salinity>=self.avoided_salinity

        self.sensing()
        ### Irradiance computation
        self.irradiance()
        ### identify the elements involved in the different scenarios
        Filter_N= self.elements.copepodid < self.elements.nauplii
        Filter_C=~Filter_N
        safe_up_salt=Normal_salt&self.elements.safe_salinity_above.astype(np.bool)
        #### need to be able to deal with false arrays of Filter_N and C...
        ### They generate empty arrays
        light_mig_N= safe_up_salt&Filter_N&(self.elements.light \
                    > self.Nauplii_light_trigger)

        light_mig_C= safe_up_salt&Filter_C&(self.elements.light \
                    > self.Copepodid_light_trigger)

        up_temp_mig= safe_up_salt&~light_mig_N&~light_mig_C& \
                    (self.elements.temperature_above \
                    > self.environment.sea_water_temperature)
        down_temp_mig=Normal_salt&~up_temp_mig&~light_mig_N& ~light_mig_C& \
                        (self.elements.temperature_below> \
                        self.environment.sea_water_temperature)

        ### Active migration
        going_down=np.logical_or(Avoiding,down_temp_mig)
        going_up=np.logical_or(np.logical_or(light_mig_N,light_mig_C),up_temp_mig)
        logger.debug("{} going down, {} going up".format(going_down.sum(), going_up.sum()))
        self.elements.z[going_down] -= self.vertical_migration_speed
        self.elements.z[going_up] +=   self.vertical_migration_speed


    def update(self):
        self.SI_pop()
        self.degree_days()
        self.advect_ocean_current()        
        self.vertical_mixing()
        self.Lice_vertical_migration()
        self.depth_test()
